
from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestPermutation(unittest.TestCase):
    @given(strategies.permutations())
    def test_pickle(self, perm):
        self.assertEqual(perm, pickle.loads(pickle.dumps(perm)))
    
    @given(st.data())
    def test_hash(self, data):
        perm1 = data.draw(strategies.permutations())
        perm2 = data.draw(strategies.permutations(len(perm1)))
        self.assertTrue(hash(perm1) != hash(perm2) or perm1 == perm2)
    
    @given(strategies.permutations())
    def test_from_index(self, perm):
        self.assertEqual(perm, curver.kernel.Permutation.from_index(len(perm), perm.index()))
    
    @given(strategies.permutations())
    def test_equal(self, perm):
        self.assertEqual(perm, perm)
    
    @given(st.data())
    def test_inverse(self, data):
        perm1 = data.draw(strategies.permutations())
        perm2 = data.draw(strategies.permutations(len(perm1)))
        identity = curver.kernel.Permutation.from_index(len(perm1), 0)
        self.assertEqual(~(~perm1), perm1)
        self.assertEqual(perm1 * ~perm1, identity)
        self.assertEqual(~perm1 * perm1, identity)
        self.assertEqual(~(perm1 * perm2), ~perm2 * ~perm1)
    
    @given(strategies.permutations())
    def test_involution(self, perm):
        identity = curver.kernel.Permutation.from_index(len(perm), 0)
        self.assertTrue(perm.order() <= 2 or perm != ~perm)
    
    @given(st.data())
    def test_even(self, data):
        perm1 = data.draw(strategies.permutations())
        perm2 = data.draw(strategies.permutations(len(perm1)))
        self.assertEqual(perm1.is_even() == perm2.is_even(), (perm1 * perm2).is_even())
        self.assertEqual(perm1.is_even() == perm2.is_even(), (perm2 * perm1).is_even())
    
    @given(strategies.permutations())
    def test_order(self, perm):
        identity = curver.kernel.Permutation.from_index(len(perm), 0)
        for i in range(1, perm.order()):
            self.assertNotEqual(perm**i, identity)
        self.assertEqual(perm**perm.order(), identity)





import math

EPSILON = 10**-8

class Vector2(object):
    # Warning: This needs to be updated if the interals of this class ever change.
    __slots__ = ['x', 'y']
    def __init__(self, x, y):
        self.x, self.y = x, y
    def __str__(self):
        return "(%s, %s)" % (self.x, self.y)
    def __repr__(self):
        return str(self)
    def __reduce__(self):
        # Having __slots__ means we need to pickle manually.
        return (self.__class__, self.to_tuple())
    def approx(self, other, epsilon=EPSILON):
        return (self - other).norm2() < epsilon
    def __eq__(self, other):
        raise TypeError('Susceptible to FPE.')
        return self.x == other.x and self.y == other.y
    def __ne__(self, other):
        return not (self == other)
    def __iter__(self):
        return iter([self.x, self.y])
    def __getitem__(self, item):
        if item == 0: return self.x
        if item == 1: return self.y
    def __setitem__(self, item, value):
        if item == 0: self.x = value
        if item == 1: self.y = value
    def __neg__(self):
        return Vector2(-self.x, -self.y)
    def __add__(self, other):
        return Vector2(self.x + other.x, self.y + other.y)
    def __radd__(self, other):
        return self + other
    def __sub__(self, other):
        return self + -other
    def __mul__(self, other):
        return Vector2(self.x * other, self.y * other)
    def __rmul__(self, other):
        return self * other
    def __div__(self, other):
        return Vector2(self.x / other, self.y / other)
    def dot(self, other):
        return self.x * other.x + self.y * other.y
    def cross(self, other):
        return self.x * other.y - self.y * other.x
    def norm2(self):
        return self.dot(self)
    def norm(self):
        return math.sqrt(self.norm2())
    def to_tuple(self):
        return (self.x, self.y)
    def rotate(self, angle):
        # Angle is in degrees.
        s, c = math.sin(math.pi * angle / 180.0), math.cos(math.pi * angle / 180.0)
        return Vector2(c * self.x - s * self.y, s * self.x + c * self.y)





''' A module for representing the curve complex of a surface. '''

from itertools import combinations
from math import factorial
import networkx

import curver

class CurveComplex(object):
    ''' This represents the curve complex of a surface. '''
    def __init__(self, triangulation):
        self.triangulation = triangulation
        self.zeta = self.triangulation.zeta
        assert(self.triangulation.is_connected())
        
        # Constants:
        # Universal:
        self.QUASICONVEXITY = 10  # Masur-Minsky.  # Ask Katie Vokes about this.
        self.BOUNDED_GEODESIC_IMAGE = 100  # BGIT.
        self.HYPERBOLICITY = 17  # Curve complex delta-hyperbolicity.
        # Uniform:
        self.D = (self.zeta**(2 * self.zeta * (2*self.HYPERBOLICITY + 2)**2))**2  # Bowditch bound on denominator [Bow08].
        # This explict bound comes from Theorems 6.2 & 6.3 of [Webb15] and Algorithm 3 of [BellWebb16].
        # TODO 4) Recheck this.
        self.XI = self.zeta // 2  # Actually should be 3g - 3 + n < 3g - 3 + 1.5n = zeta / 2
        self.M = 28 * factorial(self.XI) * self.HYPERBOLICITY * self.D * self.BOUNDED_GEODESIC_IMAGE
        E, n = self.triangulation.euler_characteristic, self.triangulation.num_vertices
        self.R = 18*E**2 - 30*E - 10*n  # Gadre and Tsai bound away from zero [GadreTsai].
        self.M2 = factorial(self.XI) * self.BOUNDED_GEODESIC_IMAGE * self.R
    
    def quasiconvex(self, a, b):
        ''' Return a polynomial-sized K--quasiconvex subset of the curve complex that contains a and b.
        
        From Algorithm 1 of [BellWebb16]_. '''
        
        # Uses Masur--Minsky theorem.
        
        assert(isinstance(a, curver.kernel.Curve))
        assert(isinstance(b, curver.kernel.Curve))
        assert(a.triangulation == self.triangulation)
        assert(b.triangulation == self.triangulation)
        
        short, conjugator = a.shorten()
        
        train_track = conjugator(b).train_track()
        _, conjugator_tt = train_track.shorten()
        
        quasiconvex = set()
        for i in range(len(conjugator_tt)):
            prefix = conjugator_tt[i:]  # Get the first bit of tt_conjugator.
            split_train_track = prefix(train_track)
            vertex_cycle = split_train_track.vertex_cycle()
            train_track_curve = prefix.inverse()(vertex_cycle)  # Pull the cycle back to the train_track.
            # Project the train track curve back to short.triangulation.
            curve = curver.kernel.Curve(short.triangulation, train_track_curve.geometric[:self.zeta])
            
            quasiconvex.add(conjugator.inverse()(curve))
        
        # Hmmm, can we skip this at only the cost of increasing some constants later?
        vertex_cycle = train_track.vertex_cycle()
        # Project the train track curve back to short.triangulation.
        curve = curver.kernel.Curve(short.triangulation, vertex_cycle.geometric[:self.zeta])
        quasiconvex.add(conjugator.inverse()(curve))
        
        return quasiconvex
    
    def tight_paths(self, a, b, length):
        ''' Return the set of all tight paths from a to b that are of the given length.
        
        From Algorithm 3 of [BellWebb16]_. '''
        
        assert(isinstance(a, curver.kernel.MultiCurve))
        assert(isinstance(b, curver.kernel.MultiCurve))
        assert(a.triangulation == self.triangulation)
        assert(b.triangulation == self.triangulation)
        assert(length >= 0)
        
        if length == 0:
            return set([(a,)]) if a == b else set()
        elif length == 1:
            return set([(a, b)]) if a.intersection(b) == 0 and a.no_common_component(b) else set()
        elif length == 2:
            m = a.boundary_union(b)  # m = \partial N(a \cup b).
            return set([(a, m, b)]) if not m.is_peripheral() and a.no_common_component(m) and b.no_common_component(m) else set()
        else:  # length >= 3.
            if not a.fills_with(b): return set()
            crush = a.crush()
            lift = crush.inverse()
            b_prime = crush(b)
            A_1 = set()
            for triangulation in b_prime.explore_ball(2*self.zeta*length + 2*self.zeta):
                for submultiarc in triangulation.sublaminations():
                    m_prime = submultiarc.boundary()
                    m = lift(m_prime)
                    A_1.add(m)
            
            P = set()
            for a_1 in A_1:
                for multipath in self.tight_paths(a_1, b, length-1):  # Recurse.
                    a_2 = multipath[0]
                    if a.boundary_union(a_2) == a_1:  # (a,) + multipath is tight:
                        P.add((a,) + multipath)
            
            return P
    
    def all_tight_geodesic_multicurves(self, a, b):
        ''' Return a set that contains all multicurves in any tight geodesic from a to b.
        
        From the first half of Algorithm 4 of [BellWebb16]_. '''
        
        assert(isinstance(a, curver.kernel.Curve))
        assert(isinstance(b, curver.kernel.Curve))
        assert(a.triangulation == self.triangulation)
        assert(b.triangulation == self.triangulation)
        
        guide = self.quasiconvex(a, b)  # U.
        L = 6*self.QUASICONVEXITY + 2  # See [Webb15].
        return set(multicurve for length in range(L+1) for c1 in guide for c2 in guide for path in self.tight_paths(c1, c2, length) for multicurve in path)
    
    def tight_geodesic(self, a, b):
        ''' Return a tight geodesic in the (multi)curve complex from a to b.
        
        From the second half of Algorithm 4 of [BellWebb16]_. '''
        
        assert(isinstance(a, curver.kernel.Curve))
        assert(isinstance(b, curver.kernel.Curve))
        assert(a.triangulation == self.triangulation)
        assert(b.triangulation == self.triangulation)
        
        # Build graph.
        vertices = list(self.all_tight_geodesic_multicurves(a, b))
        edges = [(u, v) for u, v in combinations(vertices, r=2) if u.intersection(v) == 0 and u.no_common_component(v)]
        G = networkx.Graph(edges)
        
        geodesic = networkx.shortest_path(G, a, b)  # Find a geodesic from self to other, however this might not be tight.
        
        for i in range(1, len(geodesic)-1):
            geodesic[i] = geodesic[i-1].boundary_union(geodesic[i+1])  # Tighten.
        
        return tuple(geodesic)
    
    def geodesic(self, a, b):
        ''' Return a geodesic in the curve complex from a to b.
        
        The geodesic will always come from a tight geodesic.
        From Algorithm 5 of [BellWebb16]_. '''
        
        assert(isinstance(a, curver.kernel.Curve))
        assert(isinstance(b, curver.kernel.Curve))
        assert(a.triangulation == self.triangulation)
        assert(b.triangulation == self.triangulation)
        
        return tuple(multicurve.peek_component() for multicurve in self.tight_geodesic(a, b))
    
    def distance(self, a, b):
        ''' Return the distance from a to b in the curve complex. '''
        
        # Could use self.tight_geodesic(a, b).
        return len(self.geodesic(a, b)) - 1





''' A module for representing (multi)arcs on triangulations. '''

import curver
from curver.kernel.lamination import Shortenable  # Special import needed for subclassing.
from curver.kernel.utilities import memoize  # Special import needed for decorating.

class MultiArc(Shortenable):
    ''' A Lamination in which every component is an Arc. '''
    def is_multicurve(self):
        return False
    def is_multiarc(self):
        return True
    
    def shorten_strategy(self, edge):
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        if not self.triangulation.is_flippable(edge):
            return 0
        if self.dual_weight(edge) < 0:
            return 1
        
        return 0
    
    def boundary(self):
        ''' Return the multicurve which is the boundary of a regular neighbourhood of this multiarc. '''
        
        short, conjugator = self.shorten()
        # short is a subset of the edges of the triangulation it is defined on.
        # So its geometric vector is non-positive.
        
        used = set(edge for edge in short.triangulation.edges if short(edge) < 0)
        
        # Also use any edge that is in a triangle where two sides are used.
        # Note that this does not change the boundary.
        to_check = [triangle for triangle in short.triangulation if sum(1 for edge in triangle if edge in used) == 2]
        while to_check:
            triangle = to_check.pop()
            for edge in triangle:
                if edge not in used:
                    used.add(edge)
                    used.add(~edge)
                    neighbour = short.triangulation.triangle_lookup[~edge.label]
                    if sum(1 for edge in neighbour if edge in used) == 2:
                        to_check.append(neighbour)
                    break
        
        # Now build each component by walking around the outside of the used edges.
        passed_through = set()
        geometric = [0] * short.zeta
        for edge in short.triangulation.edges:
            corner = short.triangulation.corner_lookup[edge.label]
            if edge not in passed_through and edge not in used and corner[2] in used:  # Start on a good edge.
                while edge not in passed_through:
                    geometric[edge.index] += 1
                    passed_through.add(edge)
                    corner = short.triangulation.corner_lookup[edge.label]
                    if ~corner[2] not in used:
                        edge = ~corner[2]
                    else:
                        edge = ~corner[1]
        
        boundary = short.triangulation(geometric)
        return conjugator.inverse()(boundary)
    
    def is_polygonalisation(self):
        ''' Return if this MultiArc is a polygonalisation, that is, if it cuts the surface into polygons. '''
        short, _ = self.shorten()
        
        avoid = set(index for index in short.triangulation.indices if short(index) < 0)  # All of the edges used.
        dual_tree = short.triangulation.dual_tree(avoid=avoid)
        
        return all(dual_tree[index] or index in avoid for index in short.triangulation.indices)
    
    def explore_ball(self, radius):
        ''' Extend this MultiArc to a triangulation and return all triangulations within the ball of the given radius of that one.
        
        Runs in exp(radius) time.
        Note that this is only well-defined if this multiarc is filling. '''
        
        assert(self.is_filling())
        
        short, conjugator = self.shorten()
        
        triangulations = set()
        for encoding in short.triangulation.all_encodings(radius):
            T = encoding.target_triangulation.as_lamination()
            triangulations.add(conjugator.inverse()(encoding.inverse()(T)))
        
        return triangulations
    
    def topological_type(self):
        ''' Return the topological type of this multiarc.
        
        Two multiarcs are in the same mapping class group orbit if and only their topological types are equal.
        These are labelled graphs and so equal means 'label isomorphic', so we return a custom class that uses networkx.is_isomorphic to determine equality. '''
        
        return NotImplemented  # TODO: 2) Implement.

class Arc(MultiArc):
    ''' A MultiArc with a single component. '''
    @memoize(fast=True)
    def components(self):
        return {self: 1}
    
    def parallel(self):
        ''' Return an edge that this arc is parallel to. '''
        assert(self.is_short())
        
        return min([edge for edge in self.triangulation.edges if self(edge) < 0], key=lambda e: e.label)
    
    def vertices(self):
        ''' Return the pair of vertices that this arc connects from / to. '''
        
        vertices = []
        for vertex in self.triangulation.vertices:
            for edge in vertex:
                if self(edge) < 0: vertices.append(vertex)
                if self.side_weight(edge) == -1: vertices.append(vertex)
                if self.side_weight(edge) == -2: vertices += [vertex, vertex]
        assert(len(vertices) == 2)
        return vertices
    
    def connects_distinct_vertices(self):
        ''' Return whether this arc connects between distict vertices of its underlying triangulation. '''
        
        return len(set(self.vertices())) == 2
    
    def encode_halftwist(self, power=1):
        ''' Return an Encoding of a right half twist about a regular neighbourhood of this arc, raised to the given power.
        
        Assumes that this arc connects between distinct vertices. '''
        
        if not self.connects_distinct_vertices():  # Check where it connects.
            raise curver.AssumptionError('Arc connects a vertex to itself.')
        
        short, conjugator = self.shorten()
        
        return conjugator.inverse() * curver.kernel.HalfTwist(short, power).encode() * conjugator
    
    def intersection(self, lamination):
        ''' Return the geometric intersection between self and the given lamination. '''
        
        assert(isinstance(lamination, curver.kernel.Lamination))
        assert(lamination.triangulation == self.triangulation)
        
        short, conjugator = self.shorten()
        short_lamination = conjugator(lamination)
        
        arc = short.parallel()
        
        return max(short_lamination(arc), 0)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

from collections import Counter

class TestCrush(unittest.TestCase):
    @pytest.mark.skip('Crush / Lift are not pickleable since they do not implement .package().')
    @given(strategies.curves())
    def test_pickle(self, curve):
        crush = curve.crush()
        self.assertEqual(crush, pickle.loads(pickle.dumps(crush)))
    
    @given(strategies.curves())
    @settings(max_examples=10)
    def test_inverse(self, curve):
        crush = curve.crush()
        self.assertEqual(crush, ~(~crush))
    
    @given(strategies.curves())
    @settings(max_examples=50)
    def test_lift(self, curve):
        crush = curve.crush()
        lift = crush.inverse()
        
        peripheral_curves = [lift.target_triangulation.curve_from_cut_sequence(vertex) for vertex in lift.target_triangulation.vertices]
        lifted_peripherals = [lift(lift.source_triangulation.curve_from_cut_sequence(vertex)) for vertex in lift.source_triangulation.vertices]
        self.assertEqual(Counter(lifted_peripherals), Counter(peripheral_curves + ([] if curve.is_peripheral() else [curve, curve])))
    
    @given(st.data())
    @settings(max_examples=20)
    def test_twist(self, data):
        curve = data.draw(strategies.curves())
        lamination = data.draw(strategies.curves(curve.triangulation))
        
        crush = curve.crush()
        power = data.draw(st.integers())
        twist_i = curve.encode_twist(power)
        
        self.assertEqual(crush(lamination), crush(twist_i(lamination)))





import os
from hypothesis import settings, Verbosity, HealthCheck
import pytest

def pytest_addoption(parser):
    parser.addoption('--runslow', action='store_true', help='run slow tests')

def pytest_collection_modifyitems(config, items):
    if not config.getoption('--runslow'):
        skip_slow = pytest.mark.skip(reason='need --runslow option to run')
        for item in items:
            if 'slow' in item.keywords:
                item.add_marker(skip_slow)


settings.register_profile('ci', settings(deadline=None, suppress_health_check=(HealthCheck.too_slow,)))
settings.register_profile('dev', settings(deadline=None, suppress_health_check=(HealthCheck.too_slow,), max_examples=5))
settings.register_profile('debug', settings(deadline=None, suppress_health_check=(HealthCheck.too_slow,), max_examples=5, verbosity=Verbosity.verbose))
settings.load_profile(os.getenv(u'HYPOTHESIS_PROFILE', 'dev'))





''' A module for representing triangulations along with laminations and mapping classes on them. '''

from random import choice
import re

import curver

REGEX_IS_NAME = re.compile(r'[a-z]\w*$')

class MappingClassGroup(object):
    ''' This represents a triangulation along with a collection of named mapping classes on it.
    
    It can also be given a list of curves and arcs, in which case the twists and half-twists about
    these are also added to the list of known mapping classes.
    Most importantly this object can construct a mapping class from a string descriptor.
    See self.mapping_class for additional information. '''
    def __init__(self, pos_mapping_classes=None, curves=None, arcs=None):
        if pos_mapping_classes is None: pos_mapping_classes = dict()
        if curves is None: curves = dict()
        if arcs is None: arcs = dict()
        
        for name, arc in arcs.items():
            assert(name not in pos_mapping_classes)
            pos_mapping_classes[name] = arc.encode_halftwist()
        for name, curve in curves.items():
            assert(name not in pos_mapping_classes)
            pos_mapping_classes[name] = curve.encode_twist()
        
        assert(pos_mapping_classes)
        
        if not isinstance(pos_mapping_classes, dict):
            pos_mapping_classes = dict(curver.kernel.utilities.name_objects(pos_mapping_classes))
        
        self.triangulation = list(pos_mapping_classes.values())[0].source_triangulation
        self.zeta = self.triangulation.zeta
        
        assert(all(isinstance(key, str) for key in pos_mapping_classes))
        assert(all(isinstance(pos_mapping_class, curver.kernel.MappingClass) for pos_mapping_class in pos_mapping_classes.values()))
        assert(all(pos_mapping_class.source_triangulation == self.triangulation for pos_mapping_class in pos_mapping_classes.values()))
        assert(all(REGEX_IS_NAME.match(name) for name in pos_mapping_classes))
        
        self.pos_mapping_classes = dict(pos_mapping_classes)
        self.neg_mapping_classes = dict((name.swapcase(), pos_mapping_class.inverse()) for name, pos_mapping_class in self.pos_mapping_classes.items())
        self.mapping_classes = dict(list(self.pos_mapping_classes.items()) + list(self.neg_mapping_classes.items()))
        
        self.arcs = arcs
        self.curves = curves
    
    def __repr__(self):
        return str(self)
    def __str__(self):
        pos_keys = sorted(self.pos_mapping_classes.keys(), key=curver.kernel.utilities.alphanum_key)
        return 'Mapping class group < %s > on %s' % (', '.join(pos_keys), self.triangulation)
    
    def __eq__(self, other):
        return self.triangulation == other.triangulation and self.mapping_classes == other.mapping_classes
    def __ne__(self, other):
        return not (self == other)
    
    def __getitem__(self, item):
        return self.mapping_classes[item]
    def __iter__(self):
        return iter(self.mapping_classes)
    
    def random_word(self, length, positive=True, negative=True, letters=None):
        ''' Return a random word of the required length.
        
        The letters to choose from can be specified or, alternatively, the set
        of positive, negative or all (default) mapping classes can be used by using the
        flags postive and negative. '''
        
        if letters is None:
            pos_keys = sorted(self.pos_mapping_classes.keys())
            neg_keys = sorted(self.neg_mapping_classes.keys())
            
            if positive and negative:
                letters = pos_keys + neg_keys
            elif positive and not negative:
                letters = pos_keys
            elif not positive and negative:
                letters = neg_keys
            else:
                raise TypeError('At least one of positive and negative must be allowed.')
        
        return '.'.join(choice(letters) for _ in range(length))
    
    def decompose_word(self, word):
        ''' Return a list of mapping_classes keys whose concatenation is word and the keys are chosen greedly.
        
        Raises a TypeError if the greedy decomposition fails. '''
        
        assert(isinstance(word, str))
        
        # By sorting the available keys, longest first, we ensure that any time we
        # get a match it is as long as possible.
        available_letters = sorted(self.mapping_classes, key=len, reverse=True)
        decomposition = []
        for subword in word.split('.'):
            while subword:
                for letter in available_letters:
                    if subword.startswith(letter):
                        decomposition.append(letter)
                        subword = subword[len(letter):]
                        break
                else:
                    raise TypeError('After extracting %s, the remaining %s cannot be greedly decomposed as a concatination of self.mapping_classes.' % ('.'.join(decomposition), subword))
        
        return decomposition
    
    def mapping_class(self, word):
        ''' Return the mapping class corresponding to the given word or a random one of given length if given an integer.
        
        The given word is decomposed using self.decompose_word and the composition
        of the mapping classes involved is returned.
        
        Raises a TypeError if the word does not correspond to a mapping class. '''
        
        if not isinstance(word, str):
            word = self.random_word(word)
        
        # Remove any whitespace.
        word = word.replace(' ', '')
        
        # Check for balanced parentheses.
        counter = 0
        for letter in word:
            if letter == '(': counter += 1
            if letter == ')': counter -= 1
            if counter < 0: raise TypeError('Unbalanced parentheses.')
        if counter != 0: raise TypeError('Unbalanced parentheses.')
        
        # Expand out parenthesis powers.
        # This can fail with a TypeError.
        old_word = None
        while word != old_word:  # While a change was made.
            old_word = word
            for subword, power in re.findall(r'(\([\w_\.]*\))\^(-?\d+)', word):
                decompose = self.decompose_word(subword[1:-1])
                int_power = int(power)
                if int_power > 0:
                    replacement = '.'.join(decompose) * int_power
                else:
                    replacement = '.'.join(letter.swapcase() for letter in decompose[::-1]) * abs(int_power)
                word = word.replace(subword + '^' + power, replacement)
        
        # Remove any remaining parenthesis, these do not have a power and are treated as ^1
        word = word.replace('(', '').replace(')', '')
        
        # Expand out powers without parenthesis. Here we use a greedy algorithm and take the
        # longest mapping class occuring before the power. Note that we only do one pass and so
        # only all pure powers to be expanded once, that is 'aBBB^2^3' is not recognised.
        available_letters = sorted(self.mapping_classes, key=len, reverse=True)
        for letter in available_letters:
            for subword, power in re.findall(r'(%s)\^(-?\d+)' % letter, word):
                int_power = int(power)
                word = word.replace(subword + '^' + power, (letter if int_power > 0 else letter.swapcase()) * abs(int_power))
        
        # This can fail with a TypeError.
        sequence = [item for letter in self.decompose_word(word) for item in self.mapping_classes[letter]]
        return curver.kernel.Encoding(sequence) if sequence else self.triangulation.id_encoding()
    
    def __call__(self, word):
        ''' A shortcut for self.mapping_class(...). '''
        return self.mapping_class(word)
    
    def lamination(self, geometric):
        ''' Return a new lamination on this surface assigning the specified weight to each edge. '''
        
        return self.triangulation(geometric)





''' A module for representing laminations on Triangulations. '''

from itertools import permutations

import curver
from curver.kernel.utilities import memoize  # Special import needed for decorating.

class Lamination(object):
    ''' This represents an (integral) lamination on a triangulation.
    
    Users should create these via Triangulation(...) or Triangulation.lamination(...). '''
    def __init__(self, triangulation, geometric):
        assert(isinstance(triangulation, curver.kernel.Triangulation))
        
        self.triangulation = triangulation
        self.zeta = self.triangulation.zeta
        self.geometric = geometric
        
        # Store some additional weights that are often used.
        self._dual = dict()
        self._side = dict()
        for triangle in self.triangulation:
            i, j, k = triangle  # Edges.
            a, b, c = self.geometric[i.index], self.geometric[j.index], self.geometric[k.index]
            a, b, c = max(a, 0), max(b, 0), max(c, 0)  # Correct for negatives.
            correction = min(a + b - c, b + c - a, c + a - b, 0)
            assert((a + b + c + correction) % 2 == 0)
            self._dual[i] = self._side[k] = (b + c - a + correction) // 2
            self._dual[j] = self._side[i] = (c + a - b + correction) // 2
            self._dual[k] = self._side[j] = (a + b - c + correction) // 2
    
    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.triangulation, self.geometric)
    def __str__(self):
        return '%s %s on %s' % (self.__class__.__name__, self.geometric, self.triangulation)
    def __iter__(self):
        return iter(self.geometric)
    def __call__(self, edge):
        ''' Return the geometric measure assigned to item. '''
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self.geometric[edge.index]
    def __eq__(self, other):
        return self.triangulation == other.triangulation and self.geometric == other.geometric
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        return hash(tuple(self.geometric))
    def __add__(self, other):
        # Haken sum.
        if isinstance(other, Lamination):
            if other.triangulation != self.triangulation:
                raise ValueError('Laminations must be on the same triangulation to add them.')
            
            geometric = [x + y for x, y in zip(self.geometric, other.geometric)]
            return self.triangulation(geometric)  # Have to promote.
        else:
            return NotImplemented
    def __radd__(self, other):
        return self + other  # Commutative.
    def __mul__(self, other):
        assert(isinstance(other, curver.IntegerType))
        assert(other >= 0)
        
        if other == 0:
            new_class = Lamination
        elif other == 1:
            new_class = self.__class__
        elif isinstance(other, curver.kernel.MultiArc):  # or Arc.
            new_class = curver.kernel.MultiArc
        elif isinstance(other, curver.kernel.MultiCurve):  # or Curve.
            new_class = curver.kernel.MultiCurve
        else:
            new_class = Lamination
        
        geometric = [other * x for x in self]
        # TODO: 3) Could save components if they have already been computed.
        return new_class(self.triangulation, geometric)  # Preserve promotion.
    def __rmul__(self, other):
        return self * other  # Commutative.
    
    def weight(self):
        ''' Return the geometric intersection of this lamination with its underlying triangulation. '''
        
        return sum(max(weight, 0) for weight in self)
    
    def dual_weight(self, edge):
        ''' Return the number of component of this lamination dual to the given edge.
        
        Note that when there is a terminal normal arc then we record this weight with a negative sign. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self._dual[edge]
    
    def side_weight(self, edge):
        ''' Return the number of component of this lamination dual to the given edge.
        
        Note that when there is a terminal normal arc then we record this weight with a negative sign. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self._side[edge]
    
    def is_empty(self):
        ''' Return if this lamination has no components. '''
        
        return not any(self)  # self.num_components() == 0
    
    def is_multicurve(self):
        ''' Return if this lamination is actually a multicurve. '''
        
        return not self.is_empty() and all(isinstance(component, curver.kernel.Curve) for component in self.components())
    
    def is_curve(self):
        ''' Return if this lamination is actually a curve. '''
        
        return self.is_multicurve() and self.num_components() == 1
    
    def is_multiarc(self):
        ''' Return if this lamination is actually a multiarc. '''
        
        return not self.is_empty() and all(isinstance(component, curver.kernel.Arc) for component in self.components())
    
    def is_arc(self):
        ''' Return if this lamination is actually a multiarc. '''
        
        return self.is_multiarc() and self.num_components() == 1
    
    def is_peripheral(self):
        ''' Return whether this lamination consists entirely of parallel components. '''
        
        return self.peripheral() == self
    
    def promote(self):
        ''' Return this lamination in its finest form. '''
        
        if self.is_multicurve():
            if self.is_curve():
                other = curver.kernel.Curve(self.triangulation, self.geometric)
            else:
                other = curver.kernel.MultiCurve(self.triangulation, self.geometric)
        elif self.is_multiarc():
            if self.is_arc():
                other = curver.kernel.Arc(self.triangulation, self.geometric)
            else:
                other = curver.kernel.MultiArc(self.triangulation, self.geometric)
        else:
            other = self
        
        # Move cache across.
        try:
            other._cache = self._cache  #pylint: disable=attribute-defined-outside-init
        except AttributeError:
            pass  # No cache.
        
        return other
    
    def skeleton(self):
        ''' Return the lamination obtained by collapsing parallel components. '''
        
        return self.triangulation.sum(self.components())
    
    def peek_component(self):
        ''' Return one component of this Lamination. '''
        
        return next(iter(self.components()))
    
    def intersection(self, lamination):
        ''' Return the geometric intersection number between this lamination and the given one. '''
        
        assert(isinstance(lamination, Lamination))
        
        return sum(multiplicity * component.intersection(lamination) for component, multiplicity in self.components().items())
    
    def no_common_component(self, lamination):
        ''' Return that self does not share any components with the given Lamination. '''
        
        assert(isinstance(lamination, Lamination))
        
        self_components = self.components()
        return not any(component in self_components for component in lamination.components())
    
    def train_track(self):
        ''' Return the train track underlying this lamination. '''
        # In each triangle where this lamination looks like:
        # We introduce new edges to subdivide a triangle (p, q, r) as follows:
        #            #                         #
        #           / \                       /^\
        #          /   \                     / | \
        #         /     \                   /  |  \
        #        /-------\                 /   |s(i)
        #       /         \     ===>>     /    |    \
        #      /\         /\           r /    / \    \ q
        #     /  \       /  \           /   /     \   \
        #    /    |     |    \         /  /t(j) u(k)\  \
        #   /     |     |     \       /</             \>\
        #  #-------------------#     #-------------------#
        #                                      p
        # So that afterwards every complementary region can reach a vertex.
        
        geometric = list(self.geometric)
        triangles = []
        num_subdivided = 0  # Number of subdivided triangles.
        for triangle in self.triangulation:
            dual_weights = [self.dual_weight(label) for label in triangle.labels]
            if all(weight > 0 for weight in dual_weights):  # Type 3).
                p, q, r = [curver.kernel.Edge(label) for label in triangle.labels]
                s, t, u = [curver.kernel.Edge(i) for i in range(self.zeta + 3*num_subdivided, self.zeta + 3*num_subdivided + 3)]
                triangles.append(curver.kernel.Triangle([p, ~u, t]))
                triangles.append(curver.kernel.Triangle([q, ~s, u]))
                triangles.append(curver.kernel.Triangle([r, ~t, s]))
                num_subdivided += 1
                
                geometric.extend(dual_weights)  # Record intersections with new edges.
            else:
                p, q, r = [curver.kernel.Edge(label) for label in triangle.labels]
                triangles.append(curver.kernel.Triangle([p, q, r]))
        
        T = curver.kernel.Triangulation(triangles)
        return curver.kernel.TrainTrack(T, geometric)
    
    @memoize()
    def components(self):
        ''' Return a dictionary mapping components to their multiplicities '''
        
        components = dict()
        for component, multiplicity in self.train_track().components().items():
            # Project an Arc or Curve on T back to self.triangulation.
            if isinstance(component, curver.kernel.Curve):
                component = curver.kernel.Curve(self.triangulation, component.geometric[:self.zeta])
            elif isinstance(component, curver.kernel.Arc):
                component = curver.kernel.Arc(self.triangulation, component.geometric[:self.zeta])
            else:
                raise RuntimeError('self.train_track().components() returned a non Curve / Arc.')
            
            assert(component not in components)
            components[component] = multiplicity
        
        return components
    
    def num_components(self):
        ''' Return the total number of components. '''
        
        return sum(self.components().values())
    
    def sublaminations(self):
        ''' Return all sublaminations that appear within self. '''
        
        components = self.components()
        return [self.triangulation.sum(sub) for i in range(len(components)) for sub in permutations(components, i+1)]
    
    def peripheral(self):
        ''' Return the peripheral components of this Lamination. '''
        
        geometric = [0] * self.zeta
        for vertex in self.triangulation.vertices:
            peripheral = max(min(self.side_weight(edge) for edge in vertex), 0)
            for edge in vertex:
                geometric[edge.index] += peripheral
        
        return self.triangulation(geometric)  # Have to promote.
    
    def non_peripheral(self):
        ''' Return the non-peripheral components of this Lamination. '''
        
        geometric = list(self)
        for vertex in self.triangulation.vertices:
            peripheral = max(min(self.side_weight(edge) for edge in vertex), 0)
            for edge in vertex:
                geometric[edge.index] -= peripheral
        
        return self.triangulation(geometric)  # Have to promote.
    
    def multiarc(self):
        ''' Return the maximal MultiArc contained within this lamination. '''
        
        return self.triangulation.sum([multiplicity * component for component, multiplicity in self.components().items() if isinstance(component, curver.kernel.Arc)])
    
    def multicurve(self):
        ''' Return the maximal MultiCurve contained within this lamination. '''
        
        return self.triangulation.sum([multiplicity * component for component, multiplicity in self.components().items() if isinstance(component, curver.kernel.Curve)])
    
    def boundary(self):
        ''' Return the boundary of a regular neighbourhood of this lamination. '''
        
        if self.is_empty():
            return self
        
        return self.multiarc().boundary() + self.multicurve().boundary()
    
    def is_filling(self):
        ''' Return if this Lamination fills the surface, that is, if it intersects all curves on the surface.
        
        Note that this is equivalent to:
            - it meets every non S_{0,3} component of the surface, and
            - its boundary is peripheral.
        Furthermore, if any component of this lamination is a curve then it cannot fill. '''
        
        if any(isinstance(component, curver.kernel.Curve) for component in self.components()):
            return False
        
        for component in self.triangulation.components():
            V, E = len([vertex for vertex in self.triangulation.vertices if vertex[0] in component]), len(component) // 2
            if (V, E) != (3, 3):  # component != S_{0, 3}:
                if all(self(edge) == 0 for edge in component):
                    return False
        
        return self.boundary().is_peripheral()
    
    def fills_with(self, other):
        ''' Return whether self \\cup other fills. '''
        assert(isinstance(other, Lamination))
        
        return NotImplemented  # TODO: 2) Implement!
    
    def trace(self, edge, intersection_point, length):
        ''' Return the sequence of edges encountered by following along this lamination.
        
        We start at the given edge and intersection point for the specified number of steps.
        However we terminate early if the lamination closes up before this number of steps is completed. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        start = (edge, intersection_point)
        
        assert(0 <= intersection_point < self(edge))  # Sanity.
        dual_weights = dict((edge, self.dual_weight(edge)) for edge in self.triangulation.edges)
        edges = []
        for _ in range(length):
            x, y, z = self.triangulation.corner_lookup[~edge.label]
            if intersection_point < dual_weights[z]:  # Turn right.
                edge, intersection_point = y, intersection_point
            elif dual_weights[x] < 0 and dual_weights[z] <= intersection_point < dual_weights[z] - dual_weights[x]:  # Terminate.
                break
            else:  # Turn left.
                edge, intersection_point = z, self(z) - self(x) + intersection_point
            if (edge, intersection_point) == start:  # Closes up.
                break
            edges.append(edge)
            assert(0 <= intersection_point < self(edge))  # Sanity.
        
        return edges
    
    def topological_type(self):
        ''' Return the topological type of this lamination..
        
        Two laminations are in the same mapping class group orbit if and only their topological types are equal.
        These are labelled graphs and so equal means 'label isomorphic', so we return a custom class that uses networkx.is_isomorphic to determine equality. '''
        
        # This should generalise MultiCurve.topological_type().
        # The final code should be very similar but with some additional steps:
        #  6) Let a := crush(self.multiarc()).
        #  7) Label each vertex with the topological type of a restricted to the corresponding component.
        # Note that in order for this to work the topological type MUST be compatible with the permutation of the punctures.
        
        return NotImplemented  # TODO: 2) Implement.

class Shortenable(Lamination):
    ''' A special lamination that we can put into a canonical 'short' form. '''
    
    def is_short(self):
        ''' Return whether this lamination is already short. '''
        
        return all(self.shorten_strategy(edge) == 0 for edge in self.triangulation.edges)
    
    def shorten_strategy(self, edge):  #pylint: disable=no-self-use,unused-argument
        ''' Return a float in [0, 1] describing how good flipping this edge is for making this lamination short.
        
        The higher the score, the better this flip is for reducing weight.
        Specific laminations should implement the correct strategy for getting to the minimal weight configuration. '''
        
        return NotImplemented
    
    def generic_shorten_strategy(self, edge):
        ''' Return a float in [0, 1] describing how good flipping this edge is for making this lamination short. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        if not self.triangulation.is_flippable(edge): return 0
        
        a, b, c, d, e = self.triangulation.square(edge)
        ad, bd, cd, dd, ed = [self.dual_weight(edgy) for edgy in self.triangulation.square(edge)]
        
        if ed < 0 or (ed == 0 and ad > 0 and bd > 0):
            return 1
        
        return 0
    
    @memoize()
    def shorten(self):
        ''' Return an encoding which maps this lamination to a short one, together with its image. '''
        
        lamination = self
        conjugator = None  # We used to set conjugator = lamination.triangulation.id_encoding() but this is more efficient.
        
        # Theorem: Suppose that self is not an isolating multicurve. If self.weight() > 2*self.zeta then there is a place to split.
        # Proof: TODO.
        
        # Remark: This is part of the reason why we can shorten Curves, Multiarcs and TrainTracks but not MultiCurves.
        extra = []
        edges = set(edge for edge in lamination.triangulation.edges if lamination(edge) > 0)
        while any(lamination.dual_weight(edge) < 0 for edge in edges) or any(len([edge for edge in triangle if lamination.dual_weight(edge) > 0]) == 2 for triangle in lamination.triangulation):
            edge = curver.kernel.utilities.maximum(extra + list(edges), key=lamination.generic_shorten_strategy, upper_bound=1)
            # This edge is always flippable.
            a, b, c, d, e = lamination.triangulation.square(edge)
            
            flip = lamination.triangulation.encode_flip(edge)
            try:  # Accelerate!
                if (1 -  0.1 / lamination.zeta) * lamination.weight()  > flip(lamination).weight():  # Drop is at least 10% of average edge weight.
                    raise curver.AssumptionError('Flip made definite progress.')
                
                intersection_point = lamination.side_weight(e) if lamination.side_weight(e) > 0 else -lamination.dual_weight(a)
                trace = lamination.trace(edge, intersection_point, 2*self.zeta)
                trace = trace[:trace.index(edge)+1]  # Will raise a ValueError if edge is not in trace.
                
                curve = lamination.triangulation.lamination_from_cut_sequence(trace)
                if not isinstance(curve, curver.kernel.Curve):
                    raise ValueError
                
                slope = curve.slope(lamination)  # Will raise a curver.AssumptionError if these are disjoint.
                if -1 <= slope <= 1:  # Can't accelerate. We should probably also skip cases where slope is too close to small to be efficient.
                    raise ValueError
                else:  # slope < -1 or 1 < slope:
                    move = curve.encode_twist(power=-int(slope))  # Round towards zero.
            except (ValueError, curver.AssumptionError):
                move = flip
                extra = [c, d]
            
            conjugator = move * conjugator
            lamination = move(lamination)
            if lamination(edge) <= 0:
                edges.discard(edge)
                edges.discard(~edge)
        
        extra = []
        while not lamination.is_short():
            edge = curver.kernel.utilities.maximum(extra + list(edges), key=lamination.shorten_strategy, upper_bound=1)
            # This edge is always flippable.
            a, b, c, d, e = lamination.triangulation.square(edge)
            
            move = lamination.triangulation.encode_flip(edge)
            extra = [c, d]
            conjugator = move * conjugator
            lamination = move(lamination)
            if lamination(edge) <= 0:
                edges.discard(edge)
                edges.discard(~edge)
        
        if conjugator is None: conjugator = self.triangulation.id_encoding()  # Just in case we haven't done any moves at all.
        
        return lamination, conjugator





''' A module for representing a triangulation of a punctured surface. '''

from math import factorial
import numpy as np
from functools import total_ordering
from itertools import groupby, product
from collections import Counter, defaultdict

import curver
from curver.kernel.utilities import memoize  # Special import needed for decorating.

def norm(number):
    ''' A map taking an edges label to its index.
    
    That is, x and ~x should map to the same thing. '''
    
    return max(number, ~number)

@total_ordering
class Edge(object):
    ''' This represents an oriented edge, labelled with an integer.
    
    It is specified by its label and its inverse edge is labelled with ~its label.
    
    These are really just integers but with fancy printing and indexing set up on them. '''
    
    # Warning: This needs to be updated if the interals of this class ever change.
    __slots__ = ['label', 'index']
    
    def __init__(self, label):
        self.label = label
        self.index = norm(self.label)
    
    def __repr__(self):
        return str(self)
    def __str__(self):
        return ('' if self.sign() == +1 else '~') + str(self.index)
    def __reduce__(self):
        # Having __slots__ means we need to pickle manually.
        return (self.__class__, (self.label,))
    def __eq__(self, other):
        if isinstance(other, Edge):
            return self.label == other.label
        elif isinstance(other, curver.IntegerType):
            return self.label == other
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    def __lt__(self, other):
        if isinstance(other, Edge):
            return self.label < other.label
        elif isinstance(other, curver.IntegerType):
            return self.label < other
        else:
            return NotImplemented
    def __hash__(self):
        return hash(self.label)
    
    def __invert__(self):
        ''' Return this edge but with reversed orientation. '''
        
        return Edge(~self.label)
    
    def sign(self):
        ''' Return the sign (+/-1) of this edge. '''
        
        return +1 if self.label == self.index else -1

class Triangle(object):
    ''' This represents a triangle.
    
    It is specified by a list of three edges, ordered anticlockwise.
    It builds its corners automatically. '''
    
    # Warning: This needs to be updated if the interals of this class ever change.
    __slots__ = ['edges', 'labels', 'indices']
    
    def __init__(self, edges, rotate=None):
        assert(isinstance(edges, (list, tuple)))
        assert(all(isinstance(edge, Edge) for edge in edges))
        assert(len(edges) == 3)
        
        # Edges are ordered anti-clockwise. We will cyclically permute
        # these to a canonical ordering, the one where the edges are ordered
        # minimally by label.
        best_index = min(range(3), key=lambda i: edges[i].label) if rotate is None else rotate
        
        self.edges = edges[best_index:] + edges[:best_index]
        self.labels = [edge.label for edge in self]
        self.indices = [edge.index for edge in self]
    
    def __repr__(self):
        return str(self)
    def __str__(self):
        return str(tuple(self.edges))
    def __reduce__(self):
        # Having __slots__ means we need to pickle manually.
        return (self.__class__, (self.edges,))
    def __eq__(self, other):
        if isinstance(other, Triangle):
            return self.edges == other.edges
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        return hash(tuple(self.edges))
    def __len__(self):
        return 3  # This is needed for reversed(triangle) to work.
    
    # Note that this is NOT the same convention as used in pieces.
    # There iterating and index accesses return vertices.
    def __iter__(self):
        return iter(self.edges)
    
    def __getitem__(self, index):
        return self.edges[index]
    def __contains__(self, other):
        return other in self.edges

# Remark: In other places in the code you will often see L(triangulation). This is the space
# of laminations on triangulation with the coordinate system induced by the triangulation.

class Triangulation(object):
    ''' This represents a triangulation of a punctured surface.
    
    It is specified by a list of Triangles. Its edges must be numbered 0, 1, ... '''
    def __init__(self, triangles):
        # We will sort the triangles into a canonical ordering, the one where the edges are ordered
        # minimally by label. This allows for fast comparisons.
        self.triangles = sorted(triangles, key=lambda t: t.labels)
        self.num_triangles = len(self.triangles)
        self.zeta = self.num_triangles * 3 // 2  # = self.num_edges.
        self.indices = [index for index in range(self.zeta)]
        self.labels = [label for label in range(-self.zeta, self.zeta)]
        self.edges = [Edge(label) for label in self.labels]
        self.positive_edges = [Edge(index) for index in self.indices]
        
        self.triangle_lookup = dict((edge.label, triangle) for triangle in self for edge in triangle)
        self.corner_lookup = dict((edge.label, Triangle(triangle.edges, rotate=index)) for triangle in self for index, edge in enumerate(triangle))
        
        # Group the edges into vertices and ordered anti-clockwise.
        # Here two edges are in the same class iff they have the same tail.
        unused = set(self.edges)
        self.vertices = set()
        while unused:
            vertex = [min(unused)]  # Make canonical by starting at min.
            unused.discard(vertex[0])
            while True:
                neighbour = ~self.corner_lookup[vertex[-1].label][2]
                if neighbour in unused:
                    vertex.append(neighbour)
                    unused.remove(neighbour)
                else:
                    break
            
            self.vertices.add(tuple(vertex))
        
        self.vertex_lookup = dict((edge.label, vertex) for vertex in self.vertices for edge in vertex)
        
        self.num_vertices = len(self.vertices)
        
        self.euler_characteristic = -self.zeta // 3  # = V - E + F since 3F = 2E and V = 0.
        
        # Two triangualtions are the same if and only if they have the same signature.
        self.signature = [edge.label for triangle in self for edge in triangle]
    
    @classmethod
    def from_tuple(cls, *edge_labels):
        ''' Return an Triangulation from a list of triples of edge labels.
        
        Let T be an ideal triangulaton of the punctured (oriented) surface S. Orient
        and edge e of T and assign an index i(e) in 0, ..., zeta-1. Now to each
        triangle t of T associate the triple j(t) := (j(e_1), j(e_2), j(e_3)) where:
        
            - e_1, e_2, e_3 are the edges of t, ordered acording to the orientation of t, and
            - j(e) = {  i(e) if the orientation of e agrees with that of t, and
                     { ~i(e) otherwise.
        
        Here ~x := -1 - x, the two's complement of x.
        
        We may describe T by the list [j(t) for t in T]. This function reconstructs
        T from such a list.
        
        edge_labels must be a list of triples of integers and each of
        0, ..., zeta-1, ~0, ..., ~(zeta-1) must occur exactly once. '''
        
        if len(edge_labels) == 1: edge_labels = edge_labels[0]
        
        assert(isinstance(edge_labels, (list, tuple)))
        assert(all(isinstance(labels, (list, tuple)) for labels in edge_labels))
        assert(all(len(labels) == 3 for labels in edge_labels))
        assert(edge_labels)
        
        zeta = len(edge_labels) * 3 // 2
        
        # Check that each of 0, ..., zeta-1, ~0, ..., ~(zeta-1) occurs exactly once.
        flattened = set(label for labels in edge_labels for label in labels)
        for i in range(zeta):
            if i not in flattened:
                raise TypeError('Missing label %d' % i)
            if ~i not in flattened:
                raise TypeError('Missing label ~%d' % i)
        
        return cls([Triangle([Edge(label) for label in labels]) for labels in edge_labels])
    
    @classmethod
    def from_sig(cls, sig):
        ''' Return the Triangulation defined by this signature. '''
        
        zeta, index = [curver.kernel.utilities.b64decode(x) for x in sig.split('_')]
        perm = curver.kernel.Permutation.from_index(2*zeta, index)
        tuples = [(perm[i] - zeta, perm[i+1] - zeta, perm[i+2] - zeta) for i in range(0, 2*zeta, 3)]
        return cls.from_tuple(tuples)
    
    def __repr__(self):
        return str(list(self))
    def __str__(self):
        return self.sig()
    def __iter__(self):
        return iter(self.triangles)
    def __getitem__(self, index):
        return self.triangles[index]
    def package(self):
        ''' Return a small amount of info that create_triagulation can use to reconstruct this triangulation. '''
        return ([t.labels for t in self],)
    def __reduce__(self):
        # Triangulations are already pickleable but this results in a much smaller pickle.
        return (create_triangulation, (self.__class__,) + self.package())
    def __eq__(self, other):
        return self.signature == other.signature
    def __ne__(self, other):
        return not(self == other)
    def __hash__(self):
        return hash(tuple(self.signature))
    def __call__(self, geometric):
        return self.lamination(geometric)
    
    def sig(self):
        ''' Return the signature of this triangulation. '''
        
        return curver.kernel.utilities.b64encode(self.zeta) + '_' + \
            curver.kernel.utilities.b64encode(curver.kernel.Permutation([x + self.zeta for x in self.signature]).index())
    
    def surface(self):
        ''' This return the (sorted) list of (genus, #punctures) for each component of this surface. '''
        
        # List of pairs of #vertices and #edges for each component.
        VE = [(len([vertex for vertex in self.vertices if vertex[0] in component]), len(component) // 2) for component in self.components()]
        # List of pairs of genus and #vertices edges for each component.
        GV = sorted(((2 - v + e // 3) // 2, v) for v, e in VE)
        
        return GV
    
    def max_order(self):
        ''' Return the maximum order of a mapping class on this surface. '''
        
        def order(g, v):
            ''' Return the maximum order of a periodic mapping class on S_{g, v}. '''
            # These bounds follow from the 4g + 4 bound on the closed surface [FarbMarg12]
            # and the Riemann removable singularity theorem which allows us to cap off the
            # punctures when the g > 1 without affecting this bound.
            if g > 1:
                return 4*g + 2
            elif g == 1:
                return max(v, 6)
            else:  # g == 0:
                return v
        
        # List of pairs of genus and #vertices edges for each component.
        GV = self.surface()
        
        # List of pairs of orders and multiplicity.
        OM = [(order(g, v), len(list(group))) for (g, v), group in groupby(sorted(GV))]
        
        prod = 1
        for o, m in OM:
            prod *= o * factorial(m)  # Actually need o*lcm(1, 2, ..., m), but this is easier (and not significantly larger?)
        
        return prod
    
    def components(self):
        ''' Return a list of sets of the edges in each component of self. '''
        
        classes = curver.kernel.UnionFind(self.edges)
        for edge in self.edges:
            classes.union(edge, ~edge)
        for triangle in self:
            classes.union(triangle)
        
        return list(classes)
    
    def is_connected(self):
        ''' Return if this triangulation has a single component. '''
        
        return len(self.components()) == 1
    
    def dual_tree(self, avoid=None):
        ''' Return a maximal tree in 1--skeleton of the dual of this triangulation.
        
        This are given as lists of Booleans signaling if each edge is in the tree.
        Note that when this surface is disconnected this tree is actually a forest.
        To make this unique / well-defined we return the numerically first one.
        
        If avoid is provided then none of these indices will be set in the dual tree. '''
        
        if avoid is None: avoid = set()
        
        # Kruskal's algorithm.
        dual_tree = [False] * self.zeta
        classes = curver.kernel.UnionFind(self.triangles)
        for index in self.indices:
            if index not in avoid:
                a, b = self.triangle_lookup[index], self.triangle_lookup[~index]
                if classes(a) != classes(b):
                    classes.union(a, b)
                    dual_tree[index] = True
        
        return dual_tree
    
    @memoize()
    def homology_matrix(self):
        ''' Return a matrix that kills the entries of the dual tree. '''
        
        dual_tree = self.dual_tree()
        
        M = []
        for index in self.indices:
            row = [0] * self.zeta
            if dual_tree[index]:
                edge = Edge(index)
                while True:
                    corner = self.corner_lookup[edge.label]
                    edge = corner.edges[2]
                    if not dual_tree[edge.index]:
                        row[edge.index] -= edge.sign()
                    else:
                        edge = ~edge
                    if edge.label == ~index: break
            else:
                row[index] = 1
            M.append(row)
        
        return np.matrix(M).transpose()  # Transpose the matrix.
    
    def homology_basis(self):
        ''' Return a basis for H_1(S). '''
        
        return [hc for hc in self.edge_homologies() if hc.is_canonical()]
    
    def is_flippable(self, edge):
        ''' Return if the given edge is flippable.
        
        An edge is flippable if and only if it lies in two distinct triangles. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self.triangle_lookup[edge.label] != self.triangle_lookup[~edge.label]
    
    def square(self, edge):
        ''' Return the four edges around the given edge and the diagonal.
        
        The given edge must be flippable. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        assert(self.is_flippable(edge))
        
        # Given the e, return the edges a, b, c, d, e in order.
        #
        # #<----------#
        # |     a    ^^
        # |         / |
        # |  A     /  |
        # |       /   |
        # |b    e/   d|
        # |     /     |
        # |    /      |
        # |   /       |
        # |  /     B  |
        # | /         |
        # V/    c     |
        # #---------->#
        
        corner_A, corner_B = self.corner_lookup[edge.label], self.corner_lookup[~edge.label]
        return [corner_A.edges[1], corner_A.edges[2], corner_B.edges[1], corner_B.edges[2], edge]
    
    # Build new triangulations:
    def flip_edge(self, edge):
        ''' Return a new triangulation obtained by flipping the given edge.
        
        The chosen edge must be flippable. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        assert(self.is_flippable(edge))
        
        # Use the following for reference:
        # #<----------#     #-----------#
        # |     a    ^^     |\          |
        # |         / |     | \         |
        # |  A     /  |     |  \     A2 |
        # |       /   |     |   \       |
        # |b    e/   d| --> |    \e'    |
        # |     /     |     |     \     |
        # |    /      |     |      \    |
        # |   /       |     |       \   |
        # |  /     B  |     | B2     \  |
        # | /         |     |         \ |
        # V/    c     |     |          V|
        # #---------->#     #-----------#
        
        edge_map = dict((edge, Edge(edge.label)) for edge in self.edges)
        
        # Most triangles don't change.
        triangles = [Triangle([edge_map[edgy] for edgy in triangle]) for triangle in self if edge not in triangle and ~edge not in triangle]
        
        a, b, c, d, e = self.square(edge)
        
        if edge.sign() == +1:
            triangle_A2 = Triangle([edge_map[e], edge_map[d], edge_map[a]])
            triangle_B2 = Triangle([edge_map[~e], edge_map[b], edge_map[c]])
        else:  # edge.sign() == -1:
            triangle_A2 = Triangle([edge_map[~e], edge_map[d], edge_map[a]])
            triangle_B2 = Triangle([edge_map[e], edge_map[b], edge_map[c]])
        return Triangulation(triangles + [triangle_A2, triangle_B2])
    
    def all_encodings(self, num_flips):
        ''' Yield all encodings that can be made using at most the given number of flips.
        
        Runs in exp(num_flips) time. '''
        
        if num_flips == 0:
            yield self.id_encoding()
        
        # TODO: 3) Make efficient by using the fact that disjoint flips commute.
        
        for edge in self.positive_edges:
            step = self.encode_flip(edge)
            for encoding in step.target_triangulation.all_encodings(num_flips-1):
                yield encoding * step
        
        return
    
    def relabel_edges(self, label_map):
        ''' Return a new triangulation obtained by relabelling the edges according to label_map.
        
        Assumes that label_map[index] or label_map[~index] is defined for each index. '''
        
        if isinstance(label_map, (list, tuple)):
            label_map = dict(enumerate(label_map))
        else:
            label_map = dict(label_map)
        
        # Build any missing labels.
        for i in self.indices:
            if i in label_map and ~i in label_map:
                pass
            elif i not in label_map and ~i in label_map:
                label_map[i] = ~label_map[~i]
            elif i in label_map and ~i not in label_map:
                label_map[~i] = ~label_map[i]
            else:
                raise curver.AssumptionError('Missing new label for %d.' % i)
        
        edge_map = dict((edge, Edge(label_map[edge.label])) for edge in self.edges)
        triangles = [Triangle([edge_map[edge] for edge in triangle]) for triangle in self]
        
        return Triangulation(triangles)
    
    def find_isometry(self, other, label_map):
        ''' Return the isometry from this triangulation to other defined by label_map.
        
        label_map must be a dictionary mapping self.labels to other.labels. Labels may
        be omitted if they are determined by other given ones and these will be found
        automatically. Additionally, if an entire component is omitted then we assume
        that the map is the identity on it.
        
        Assumes that such an isometry exists and is unique. '''
        
        assert(isinstance(label_map, dict))
        
        # Make a local copy as we may need to make a lot of changes.
        label_map = dict(label_map)
        
        source_orders = dict([(edge.label, len(vertex)) for vertex in self.vertices for edge in vertex])
        target_orders = dict([(edge.label, len(vertex)) for vertex in other.vertices for edge in vertex])
        # We do a depth first search extending the corner map across the triangulation.
        # This is a stack of labels that may still have consequences to check.
        to_process = [(edge_from_label, label_map[edge_from_label]) for edge_from_label in label_map]
        while to_process:
            from_label, to_label = to_process.pop()
            
            neighbours = [
                (~from_label, ~to_label),
                (self.corner_lookup[from_label].labels[1], other.corner_lookup[to_label].labels[1])
                ]
            for new_from_label, new_to_label in neighbours:
                if new_from_label in label_map:
                    # Check that this map is still consistent.
                    if new_to_label != label_map[new_from_label]:
                        raise curver.AssumptionError('This label_map does not extend to an isometry.')
                else:
                    # Extend the map.
                    if source_orders[new_from_label] != target_orders[new_to_label]:
                        raise curver.AssumptionError('This label_map does not extend to an isometry.')
                    label_map[new_from_label] = new_to_label
                    to_process.append((new_from_label, new_to_label))
        
        # Now assume that the map is the identity on all unmapped edges.
        # If we have gotten this far then the and unmapped edges must be entire components.
        used_labels = set(x for key_value in label_map.items() for x in key_value)
        for edge in self.edges:
            if edge.label not in used_labels:
                if self.corner_lookup[edge.label] != other.corner_lookup[edge.label]:
                    raise curver.AssumptionError('This label_map does not extend to an isometry.')
                label_map[edge.label] = edge.label
        
        return curver.kernel.Isometry(self, other, label_map)
    
    def isometries_to(self, other):
        ''' Return a list of all isometries from this triangulation to other. '''
        
        assert(isinstance(other, Triangulation))
        
        if self.zeta != other.zeta:
            return []
        
        # TODO: 3) Make this more efficient by avoiding trying all mappings.
        
        # Isometries are determined by where a single triangle is sent.
        sources = [min(component, key=lambda edge: len(self.vertex_lookup[edge.label])) for component in self.components()]
        degrees = [len(self.vertex_lookup[edge.label]) for edge in sources]
        targets = [[edge for edge in other.edges if len(other.vertex_lookup[edge.label]) == degree] for degree in degrees]
        
        isometries = []
        for chosen_targets in product(*targets):
            try:
                isometries.append(self.find_isometry(other, dict(zip(sources, chosen_targets))))
            except curver.AssumptionError:  # Map does not extend uniquely.
                pass
        
        return isometries
    
    def self_isometries(self):
        ''' Return a list of isometries taking this triangulation to itself. '''
        
        return self.isometries_to(self)
    
    def is_isometric_to(self, other):
        ''' Return if there are any orientation preserving isometries from this triangulation to other. '''
        
        assert(isinstance(other, Triangulation))
        
        return len(self.isometries_to(other)) > 0
    
    # Laminations we can build on this triangulation.
    def lamination(self, weights):
        ''' Return a new lamination on this surface assigning the specified weight to each edge. '''
        
        assert(len(weights) == self.zeta)
        # Should check all dual weights.
        
        return curver.kernel.Lamination(self, weights).promote()
    
    def lamination_from_cut_sequence(self, sequence):
        ''' Return a new lamination on this surface based on the sequence of edges that this Curve / Arc crosses. '''
        
        count = Counter(sequence)
        return self([count[i] + count[~i] for i in range(self.zeta)])
    
    def curve_from_cut_sequence(self, sequence):
        ''' Return a new curve on this surface based on the sequence of edges that this Curve crosses.
        
        WARNING: Be extremely careful with this method since it does NOT check that this produces a curve. '''
        
        count = Counter(sequence)
        return curver.kernel.Curve(self, [count[i] + count[~i] for i in range(self.zeta)])
    
    def empty_lamination(self):
        ''' Return the empty lamination defined on this triangulation. '''
        
        return self([0] * self.zeta)
    
    def as_lamination(self):
        ''' Return this triangulation as a lamination. '''
        
        return self([-1] * self.zeta)
    
    def sum(self, laminations):
        ''' An efficient way of summing multiple laminations without computing intermediate values. '''
        
        laminations = list(laminations)
        if all(isinstance(lamination, curver.kernel.Lamination) for lamination in laminations):
            if not laminations:
                return self.empty_lamination()
            
            if any(lamination.triangulation != self for lamination in laminations):
                raise ValueError('Laminations must all be defined on this triangulation to add them.')
            
            geometric = [sum(weights) for weights in zip(*laminations)]
            return self(geometric)  # Have to promote.
        else:
            return NotImplemented
    
    def disjoint_sum(self, laminations):
        ''' An efficient way of summing multiple disjoint laminations without computing intermediate values. '''
        
        laminations = list(laminations)
        if all(isinstance(lamination, curver.kernel.Lamination) for lamination in laminations):
            if any(lamination.triangulation != self for lamination in laminations):
                raise ValueError('Laminations must all be defined on this triangulation to add them.')
            
            num_components = sum(lamination.num_components() for lamination in laminations)
            if num_components == 0:
                return self.empty_lamination()
            
            geometric = [sum(weights) for weights in zip(*laminations)]
            if all(isinstance(lamination, curver.kernel.MultiArc) for lamination in laminations):
                if num_components == 1:
                    new_class = curver.kernel.Arc
                else:  # num_components > 1:
                    new_class = curver.kernel.MultiArc
            elif all(isinstance(lamination, curver.kernel.MultiCurve) for lamination in laminations):
                if num_components == 1:
                    new_class = curver.kernel.Curve
                else:  # num_components > 1:
                    new_class = curver.kernel.MultiCurve
            else:  # Mixed.
                new_class = curver.kernel.Lamination
            
            return new_class(self, geometric)
        else:
            return NotImplemented
    
    def edge_arc(self, edge):
        ''' Return the given edge as an Arc. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return curver.kernel.Arc(self, [0 if i != edge.index else -1 for i in range(self.zeta)])  # Avoids promote.
    
    def edge_arcs(self):
        ''' Return a list containing the Arc representing each Edge.
        
        As these fill, by Alexander's trick a mapping class is the identity if and only if it fixes all of them. '''
        
        return [self.edge_arc(edge) for edge in self.positive_edges]  # Could use self.lamination.
    
    def edge_homologies(self):
        ''' Return a list containing the HomologyClass of each Edge. '''
        
        # Could skip those in self.dual_tree().
        return [curver.kernel.HomologyClass(self, [0 if i != index else 1 for i in range(self.zeta)]) for index in self.indices]
    
    def id_isometry(self):
        ''' Return the isometry representing the identity map. '''
        
        return curver.kernel.Isometry(self, self, dict((i, i) for i in self.labels))
    
    def id_encoding(self):
        ''' Return an encoding of the identity map on this triangulation. '''
        
        return self.id_isometry().encode()
    
    def encode_flip(self, edge):
        ''' Return an encoding of the effect of flipping the given edge.
        
        The given edge must be flippable. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        assert(self.is_flippable(edge))
        
        new_triangulation = self.flip_edge(edge)
        
        return curver.kernel.EdgeFlip(self, new_triangulation, edge).encode()
    
    def encode_relabel_edges(self, label_map):
        ''' Return an encoding of the effect of relabelling the edges according to label_map.
        
        Assumes that label_map[index] or label_map[~index] is defined for each index. '''
        
        if isinstance(label_map, (list, tuple)):
            label_map = dict(enumerate(label_map))
        else:
            label_map = dict(label_map)
        
        # Build any missing labels.
        # We need to repeat this code so that we can build the isometry in a minute.
        for i in self.indices:
            if i in label_map and ~i in label_map:
                pass
            elif i not in label_map and ~i in label_map:
                label_map[i] = ~label_map[~i]
            elif i in label_map and ~i not in label_map:
                label_map[~i] = ~label_map[i]
            else:
                raise curver.AssumptionError('Missing new label for %d.' % i)
        
        new_triangulation = self.relabel_edges(label_map)
        
        return curver.kernel.Isometry(self, new_triangulation, label_map).encode()
    
    def encode(self, sequence):
        ''' Return the encoding given by sequence.
        
        This consists of EdgeFlips, Isometries and LinearTransformations. Furthermore there are
        several conventions that allow these to be specified by a smaller amount of information:
        
         - An integer x represents EdgeFlip(..., edge_label=x)
         - A dictionary which has i or ~i as a key (for every i) represents a relabelling.
         - A dictionary which is missing i and ~i (for some i) represents an isometry back to this triangulation.
         - None represents the identity isometry.
        
        This sequence is read in reverse in order respect composition. For example:
        
            self.encode([1, {1: ~2}, 2, 3, ~4])
        
        is the mapping class which: flips edge ~4, then 3, then 2, then relabels
        back to the starting triangulation via the isometry which takes 1 to ~2 and
        then finally flips edge 1. '''
        
        assert(isinstance(sequence, (list, tuple)))
        assert(sequence)
        
        T = self
        h = None
        for item in reversed(sequence):
            if isinstance(item, curver.IntegerType):  # Flip.
                g = T.encode_flip(item)
            elif isinstance(item, dict):  # Isometry.
                if all(i in item or ~i in item for i in self.indices):
                    g = T.encode_relabel_edges(item)
                else:  # If some edges are missing then we assume that we must be mapping back to this triangulation.
                    g = T.find_isometry(self, item).encode()
            elif isinstance(item, tuple) and len(item) == 2:  # Twist or HalfTwist.
                label, power = item
                edge = Edge(label)
                
                # Check where it connects.
                if T.vertex_lookup[edge.label] == T.vertex_lookup[~edge.label]:  # Twist.
                    edges = curver.kernel.utilities.cyclic_slice(T.vertex_lookup[edge.label], edge, ~edge)[1:]
                    curve = T.curve_from_cut_sequence(edges)  # Avoids promote.
                    g = curver.kernel.Twist(curve, power).encode()
                else:  # HalfTwist.
                    arc = T.edge_arc(edge)
                    g = curver.kernel.HalfTwist(arc, power).encode()
            elif item is None:  # Identity isometry.
                g = T.id_encoding()
            elif isinstance(item, curver.kernel.Encoding):  # Encoding.
                g = item
            elif isinstance(item, curver.kernel.Move):  # Move.
                g = item.encode()
            else:  # Other.
                g = item.encode()
            
            if h is None:
                h = g
            else:
                h = g * h
            T = h.target_triangulation
        
        return h

def create_triangulation(cls, edge_labels):
    ''' A helper function for pickling. '''
    
    return cls.from_tuple(edge_labels)





''' A module for representing homology classes on triangulations. '''

import curver

class HomologyClass(object):
    ''' This represents a homology class of a triangulation (relative to its vertices). '''
    def __init__(self, triangulation, algebraic):
        assert(isinstance(triangulation, curver.kernel.Triangulation))
        
        self.triangulation = triangulation
        self.zeta = self.triangulation.zeta
        self.algebraic = list(algebraic)
        assert(len(self.algebraic) == self.zeta)
    
    def __repr__(self):
        return str(self)
    def __str__(self):
        return str(self.algebraic)
    def __iter__(self):
        return iter(self.algebraic)
    def __call__(self, edge):
        ''' Return the geometric measure assigned to item. '''
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self.algebraic[edge.index] * edge.sign()
    def __eq__(self, other):
        return self.triangulation == other.triangulation and self.canonical().algebraic == other.canonical().algebraic
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        return hash(tuple(self.canonical().algebraic))
    def __add__(self, other):
        if isinstance(other, HomologyClass):
            if other.triangulation != self.triangulation:
                raise ValueError('Homology classes must be on the same triangulation to add them.')
            
            algebraic = [x + y for x, y in zip(self, other)]
            return HomologyClass(self.triangulation, algebraic)
        elif other == 0:  # So we can use sum.
            return self
        else:
            return NotImplemented
    def __radd__(self, other):
        return self + other
    def __neg__(self):
        algebraic = [-x for x in self]
        return HomologyClass(self.triangulation, algebraic)
    def __sub__(self, other):
        return self + (-other)
    def canonical(self):
        ''' Return the canonical form of this HomologyClass.
        
        This is the HomologyClass that is homologous to this one and has weight 0 on each edge of the standard dual tree of the underlying triangulation. '''
        
        return HomologyClass(self.triangulation, curver.kernel.matrix_vector(self.triangulation.homology_matrix(), self.algebraic))
    
    def is_canonical(self):
        ''' Return whether this homology class is already in canonical form.
        
        This is useful for constructing a basis of H_1(S). '''
        
        return self.canonical().algebraic == self.algebraic





import re

import curver

REGEX_IS_SPHERE_BRAID = re.compile(r'SB_(?P<num_strands>\d+)$')

# Based on code by William Worden.

# TODO: 3) Document all of these cases.

def S_0_n(n):
    # A triangulation of S_{0,n}.
    assert(n >= 3)
    
    # We'll build the following triangulation of S_{0,n}:
    #
    # |      |      |      |       |
    # |0     |1     |2     |3      |n-3
    # |      |      |      |       |
    # | 2n-4 | 2n-3 | 2n-2 |       | 3n-7
    # X------X------X------X- ... -X------X
    #        |      |      |       |      |
    #        |      |      |       |      |
    #        |n-2   |n-1   |n      |2n-6  |2n-5
    #        |      |      |       |      |
    #
    # Note that the (n-1)st puncture is not shown here and is on the "boundary" of the disk.
    # There are two families of triangles on the top and the bottom and two exceptional cases:
    #  1) Top triangles are given by [i, i+2n-4, ~(i+1)] for i in range(0, n-3)],
    #  2) Bottom triangles are given by [i, ~(i+1), ~(i+n-1)] for i in range(n-2, 2n-5),
    #  3) Left triangle given by [~0, ~(n-2), ~(2n-4)], and
    #  4) Right triangle given by [n-3, 3n-7, 2n-5].
    T = curver.create_triangulation(
        [(i, i + 2 * n - 4, ~(i + 1)) for i in range(0, n - 3)] + \
        [(i, ~(i + 1), ~(i + n - 1)) for i in range(n-2, 2*n - 5)] + \
        [(~0, ~(n - 2), ~(2 * n - 4)), (n - 3, 3 * n - 7, 2 * n - 5)]
        )
    
    # We'll then create an arc connecting the ith to (i+1)st punctures.
    # Note that the arcs connecting (n-2)nd & (n-1)st and (n-1)st & 0th are special.
    
    curves, arcs = dict(), dict()
    for i in range(n-2):
        arcs['s_%d' % i] = T.edge_arc(2*n-4+i)
    arcs['s_%d' % (n-2)] = T.edge_arc(2*n-5)
    arcs['s_%d' % (n-1)] = T.edge_arc(0)
    
    return curver.kernel.MappingClassGroup(curves=curves, arcs=arcs)

def S_1_n(n):
    assert(n >= 1)
    
    curves, arcs = dict(), dict()
    if n == 1:
        T = curver.create_triangulation((0, 1, 2), (~0, ~1, ~2))
        
        curves['a_0'] = T.curve_from_cut_sequence([0, 2])
        curves['b_0'] = T.curve_from_cut_sequence([0, 1])
    else:  # n > 1:
        T = curver.create_triangulation(
            [(0, 1, 2)] + \
            [(~(1+2*i), 1+2*i+2, 1+2*i+3) for i in range(n-1)] + \
            [(2*n+1, ~(2*n), ~(2*n-1))] + \
            [(2*n+1 + i, ~(2*n-2*i), ~(2*n + i)) for i in range(1, n-1)] + \
            [(~0, ~(3*n-1), ~2)]
            )
        
        curves['a_0'] = T.curve_from_cut_sequence([0, 1] + [3 + 2*j for j in range(n-1)] + [] + [2*n+1 + j for j in range(n-1)])
        curves['b_0'] = T.curve_from_cut_sequence([0, 2])
        # The twists obtained by pushing a_0 across the punctures.
        # Note that if the loop ran with i=0 then it would create p_0 == a_0.
        for i in range(1, n):
            curves['p_%d' % i] = T.curve_from_cut_sequence([0, 1] + [3 + 2*j for j in range(n-1-i)] + [2*n + 2 - 2*i] + [2*n+i + j for j in range(n-i)])
        # The half-twists that permute the ith and (i+1)st punctures.
        arcs['s_0'] = T.edge_arc(2*n - 1)
        for i in range(1, n):
            arcs['s_%d' % i] = T.edge_arc(2*n + 2 - 2*i)
    
    return curver.kernel.MappingClassGroup(curves=curves, arcs=arcs)

def S_2_n(n):
    assert(n >= 1)
    
    curves, arcs = dict(), dict()
    if n == 1:
        T = curver.create_triangulation(
            (0, 1, 2), (~1, 3, 4), (~2, ~3, ~4),
            (~0, 5, 6), (~5, 7, 8), (~6, ~7, ~8)
            )
        
        curves['a_0'] = T.curve_from_cut_sequence([1, 2, 3])
        curves['a_1'] = T.curve_from_cut_sequence([5, 6, 7])
        curves['b_0'] = T.curve_from_cut_sequence([1, 2, 4])
        curves['b_1'] = T.curve_from_cut_sequence([5, 6, 8])
        curves['c_0'] = T.curve_from_cut_sequence([0, 1, 2, 3, 0, 5, 6, 7])
    else:  # n > 1:
        T = curver.create_triangulation(
            [(0, 1, 2), (~1, 3, 4), (~2, ~3, ~4), (~0, 5, 6), (~6, ~(3*n+5), ~8)] + \
            [(~(5+2*i), 5+2*i+2, 5+2*i+3) for i in range(n)] + \
            [(2*n+7, ~(2*n+6), ~(2*n+5))] + \
            [(2*n+7+i, ~(2*n+6 - 2*i), ~(2*n+6+i)) for i in range(1, n-1)]
            )
        
        curves['a_0'] = T.curve_from_cut_sequence([1, 2, 3])
        curves['a_1'] = T.curve_from_cut_sequence([6, 5] + [7 + j*2 for j in range(n)] + [] + [2*n+7 + j for j in range(n-1)])
        curves['b_0'] = T.curve_from_cut_sequence([1, 2, 4])
        curves['b_1'] = T.curve_from_cut_sequence([5, 6, 8])
        curves['c_0'] = T.curve_from_cut_sequence([6, 0, 2, 4, 1, 2, 3, 4, 2, 0, 5] + [7 + j*2 for j in range(n)] + [] + [2*n+7 + j for j in range(n-1)])
        # The twists obtained by pushing a_1 across the punctures.
        # Note that if the loop ran with i=0 then it would create p_0 == a_1.
        for i in range(1, n):
            curves['p_%d' % i] = T.curve_from_cut_sequence([6, 5] + [7 + j*2 for j in range(n-i)] + [2*n + 8 - 2*i] + [2*n+6+i + j for j in range(n-i)])
        # The half-twists that permute the ith and (i+1)st punctures.
        arcs['s_0'] = T.edge_arc(2*n + 5)
        for i in range(1, n):
            arcs['s_%d' % i] = T.edge_arc(2*n + 8 - 2*i)
    
    return curver.kernel.MappingClassGroup(curves=curves, arcs=arcs)


def S_3_n(n):
    assert(n >= 1)
    
    curves, arcs = dict(), dict()
    if n == 1:
        T = curver.create_triangulation(
            (0, 1, 2), (~1, 3, 4), (~2, ~3, ~4),
            (5, 6, 7), (~6, 8, 9), (~7, ~8, ~9),
            (10, 11, 12), (~11, 13, 14), (~12, ~13, ~14),
            (~0, ~5, ~10)
            )
        
        curves['a_0'] = T.curve_from_cut_sequence([1, 2, 3])
        curves['a_1'] = T.curve_from_cut_sequence([6, 7, 8])
        curves['a_2'] = T.curve_from_cut_sequence([11, 12, 13])
        curves['b_0'] = T.curve_from_cut_sequence([1, 2, 4])
        curves['b_1'] = T.curve_from_cut_sequence([6, 7, 9])
        curves['b_2'] = T.curve_from_cut_sequence([11, 12, 14])
        curves['c_0'] = T.curve_from_cut_sequence([0, 2, 4, 1, 2, 3, 4, 2, 0, 5, 6, 8, 7, 5])
        curves['c_1'] = T.curve_from_cut_sequence([5, 7, 9, 6, 7, 8, 9, 7, 5, 10, 11, 13, 12, 10])
    else:  # n > 1:
        T = curver.create_triangulation(
            [(0, 1, 2), (~1, 3, 4), (~2, ~3, ~4), (5, 6, 7), (~6, 8, 9), (~7, ~8, ~9)] + \
            [(10, 11, 12), (~11, 13, 14), (~12, ~(3*n+11), ~14)] + \
            [(~(13+2*i), 13+2*i+2, 13+2*i+3) for i in range(n-1)] + \
            [(2*n+13, ~(2*n+12), ~(2*n+11))] + \
            [(2*n+13+i, ~(2*n+12-2*i), ~(2*n+12+i)) for i in range(1, n-1)] + \
            [(~0, ~5, ~10)]
            )
        
        curves['a_0'] = T.curve_from_cut_sequence([1, 2, 3])
        curves['a_1'] = T.curve_from_cut_sequence([6, 7, 8])
        curves['a_2'] = T.curve_from_cut_sequence([12, 11] + [13 + j*2 for j in range(n)] + [] + [2*n+13 + j for j in range(n-1)])
        curves['b_0'] = T.curve_from_cut_sequence([1, 2, 4])
        curves['b_1'] = T.curve_from_cut_sequence([6, 7, 9])
        curves['b_2'] = T.curve_from_cut_sequence([11, 12, 14])
        curves['c_0'] = T.curve_from_cut_sequence([0, 2, 4, 3, 2, 1, 4, 2, 0, 5, 7, 8, 6, 5])
        curves['c_1'] = T.curve_from_cut_sequence([12, 10, 5, 7, 9, 6, 7, 8, 9, 7, 5, 10, 11] + [13 + j*2 for j in range(n)] + [] + [2*n+13 + j for j in range(n-1)])
        # The twists obtained by pushing a_2 across the punctures.
        # Note that if the loop ran with i=0 then it would create p_0 == a_2.
        for i in range(1, n):
            curves['p_%d' % i] = T.curve_from_cut_sequence([12, 11] + [13 + j*2 for j in range(n-i)] + [2*n + 14 - 2*i] + [2*n+12+i + j for j in range(n-i)])
        # The half-twists that permute the ith and (i+1)st punctures.
        arcs['s_0'] = T.edge_arc(2*n + 11)
        for i in range(1, n):
            arcs['s_%d' % i] = T.edge_arc(2*n + 14 - 2*i)
    
    return curver.kernel.MappingClassGroup(curves=curves, arcs=arcs)

def S_g_n(g, n):
    assert(g >= 4)
    assert(n >= 1)
    
    curves, arcs = dict(), dict()
    if n == 1:
        T = curver.create_triangulation(
            # Build the fins using the first 5g edges.
            [(5*i+0, 5*i+1, 5*i+2) for i in range(g)] + \
            [(~(5*i+1), 5*i+3, 5*i+4) for i in range(g)] + \
            [(~(5*i+2), ~(5*i+3), ~(5*i+4)) for i in range(g)] + \
            # Finally triangulate the centre polygon using the last g - 3 edges.
            [(~0, ~5, 5*g)] + \
            [(5*g+1+i, ~(5*g+i), ~(5*i+10)) for i in range(g-4)] + \
            [(~(5*g+g-4), ~(5*g-10), ~(5*g-5))]
            )  # 0, ..., 6g - 4.
        
        for i in range(g):
            curves['a_%d' % i] = T.curve_from_cut_sequence([5*i+1, 5*i+2, 5*i+3])
            curves['b_%d' % i] = T.curve_from_cut_sequence([5*i+1, 5*i+2, 5*i+4])
        curves['c_%d' % 0] = T.curve_from_cut_sequence([5*0+j for j in [0, 2, 4, 3, 2, 1, 4, 2, 0, 5, 6, 8, 7, 5]])
        for i in range(1, g-2):
            curves['c_%d' % i] = T.curve_from_cut_sequence([5*i+j for j in [0, 2, 4, 3, 2, 1, 4, 2, 0, 5, 6, 8, 7, 5]] + [5*g + i - 1, 5*g + i - 1])
        curves['c_%d' % (g-2)] = T.curve_from_cut_sequence([5*(g-2)+j for j in [0, 2, 4, 3, 2, 1, 4, 2, 0, 5, 6, 8, 7, 5]])
    else:  # n > 1:
        T = curver.create_triangulation(
            # Build the fins using the first 5g edges.
            [(5*i+0, 5*i+1, 5*i+2) for i in range(g)] + \
            [(~(5*i+1), 5*i+3, 5*i+4) for i in range(g)] + \
            [(~(5*i+2), ~(5*i+3), ~(5*i+4)) for i in range(g-1)] + \
            # Don't forget the last one is special.
            [(~(5*(g-1)+2), ~(5*(g-1)+4+(3*n-3)), ~(5*(g-1)+4))] + \
            # Now fold to put in the additional punctures using the next 3n - 3 edges.
            [(~(5*g-2 + 2*i), 5*g-2 + 2*i + 2, 5*g-2 + 2*i + 3)  for i in range(n-1)] + \
            [(5*g-2 + 2*(n-1) + 2, ~(5*g-2 + 2*(n-1) + 1), ~(5*g-2 + 2*(n-1)))] + \
            [(5*g-2 + 2*(n-1) + 2 + i, ~(5*g-2 + 2*(n-1) + 1 - 2*i), ~(5*g-2 + 2*(n-1) + 2 + i - 1)) for i in range(1, n-1)] + \
            # Finally triangulate the centre polygon using the last g - 3 edges.
            [(~0, ~5, 5*g+3*n-3)] + \
            [(5*g+3*n-3+1+i, ~(5*g+3*n-3+i), ~(5*i+10)) for i in range(g-4)] + \
            [(~(5*g+3*n-3+g-4), ~(5*g-10), ~(5*g-5))]
            )
        
        for i in range(g-1):
            curves['a_%d' % i] = T.curve_from_cut_sequence([5*i+1, 5*i+2, 5*i+3])
        curves['a_%d' % (g-1)] = T.curve_from_cut_sequence([5*(g-1)+1, 5*(g-1)+2] + [5*(g-1)+3 + j*2 for j in range(n)] + [] + [5*g+2*n-2 + j for j in range(n-1)])
        
        for i in range(g):
            curves['b_%d' % i] = T.curve_from_cut_sequence([5*i+1, 5*i+2, 5*i+4])
        curves['c_%d' % 0] = T.curve_from_cut_sequence([5*0+j for j in [0, 2, 4, 3, 2, 1, 4, 2, 0, 5, 6, 8, 7, 5]])
        for i in range(1, g-2):
            curves['c_%d' % i] = T.curve_from_cut_sequence([5*i+j for j in [0, 2, 4, 3, 2, 1, 4, 2, 0, 5, 6, 8, 7, 5]] + [5*g + 3*n - 3 + i - 1, 5*g + 3*n - 3 + i - 1])
        
        curves['c_%d' % (g-2)] = T.curve_from_cut_sequence([5*(g-2) + j for j in [7, 5, 0, 2, 4, 1, 2, 3, 4, 2, 0, 5, 6]] + [5*(g-1)+3 + j*2 for j in range(n)] + [] + [5*g+2*n-2 + j for j in range(n-1)])
        
        # The twists obtained by pushing a_{g-1} across the punctures.
        # Note that if the loop ran with i=0 then it would create p_0 == a_{g-1}.
        for i in range(1, n):
            curves['p_%d' % i] = T.curve_from_cut_sequence([5*(g-1)+1, 5*(g-1)+2] + [5*(g-1)+3 + j*2 for j in range(n-i)] + [5*g + 2*n - 1 - 2*i] + [5*g+2*n-3+i + j for j in range(n-i)])
        # The half-twists that permute the ith and (i+1)st punctures.
        arcs['s_0'] = T.edge_arc(5*g + 2*n - 4)
        for i in range(1, n):
            arcs['s_%d' % i] = T.edge_arc(5*g + 2*n - 1 - 2*i)
    
    return curver.kernel.MappingClassGroup(curves=curves, arcs=arcs)

def load_old(surface):
    if surface == 'S_0_4':
        S = load(0, 4)
        return curver.kernel.MappingClassGroup({
            'a': S('s_1'),
            'b': S('s_2'),
            'c': S('s_3'),
            'd': S('s_0'),
            })
    elif surface == 'S_1_1':
        S = load(1, 1)
        return curver.kernel.MappingClassGroup({
            'a': S('a_0'),
            'b': S('b_0'),
            })
    elif surface == 'S_1_2':
        S = load(1, 2)
        return curver.kernel.MappingClassGroup({
            'a': S('a_0'),
            'b': S('b_0'),
            'c': S('p_1'),
            'x': S('s_1'),
            })
    elif surface == 'S_1_2p':
        S = load(1, 2)
        return curver.kernel.MappingClassGroup({
            'a': S('a_0'),
            'b': S('b_0'),
            'c': S('p_0'),
            })
    elif surface == 'S_2_1':
        S = load(2, 1)
        return curver.kernel.MappingClassGroup({
            'a': S('b_0'),
            'b': S('c_0'),
            'c': S('b_1'),
            'd': S('a_1'),
            'e': S('(a_0.b_0.c_0)^4.A_1'),
            'f': S('a_0'),
            })
    elif surface == 'S_3_1':
        S = load(3, 1)
        return curver.kernel.MappingClassGroup({
            'a': S('b_0'),
            'b': S('c_0'),
            'c': S('b_1'),
            'd': S('c_1'),
            'e': S('b_2'),
            'f': S('(a_0.b_0.c_0.b_1.c_1)^6.A_2'),
            'g': S('a_2'),
            'h': S('a_1'),
            })
    elif surface == 'S_4_1':
        S = load(4, 1)
        return curver.kernel.MappingClassGroup({
            'a': S('b_0'),
            'b': S('c_0'),
            'c': S('b_1'),
            'd': S('c_1'),
            'e': S('b_2'),
            'f': S('c_2'),
            'g': S('b_3'),
            'h': S('(a_0.b_0.c_0.b_1.c_1.b_2.c_2)^8.A_3'),
            'i': S('a_3'),
            'j': S('a_2'),
            })
    elif surface == 'S_5_1':
        S = load(5, 1)
        return curver.kernel.MappingClassGroup({
            'a': S('b_0'),
            'b': S('c_0'),
            'c': S('b_1'),
            'd': S('c_1'),
            'e': S('b_2'),
            'f': S('c_2'),
            'g': S('b_3'),
            'h': S('c_3'),
            'i': S('b_4'),
            'j': S('(a_0.b_0.c_0.b_1.c_1.b_2.c_2.b_3.c_3)^10.A_4'),
            'k': S('a_4'),
            'l': S('a_3'),
            })
    elif REGEX_IS_SPHERE_BRAID.match(surface):
        return S_0_n(int(REGEX_IS_SPHERE_BRAID.match(surface).groupdict()['num_strands']))
    else:
        raise ValueError('Unknown surface: %s' % surface)

def load(*args):
    ''' Return the requested example MappingClassGroup.
    
    The mapping class group can either be specified by:
    
        - a pair (g, n) in which case the Lickorish generating set is returned (see Figures 4.5 and 4.10 of [FarbMarg12]), or
        - a string 'S_g_n' in which case the corresponding flipper / Twister generating set is returned.
    '''
    
    if len(args) == 1:  # Load an old flipper surface.
        surface = args[0]
        return load_old(surface)
    elif len(args) == 2:  # Build Mod(S_{g, n}).
        g, n = args
        assert(isinstance(g, curver.IntegerType))
        assert(isinstance(n, curver.IntegerType))
        
        zeta = 6*g + 3*n - 6
        if g < 0 or n < 0 or zeta < 3:
            raise curver.AssumptionError('Surface cannot be triangulated.')
        
        if g == 0:  # and n >= 3.
            return S_0_n(n)
        elif g == 1:
            return S_1_n(n)
        elif g == 2:
            return S_2_n(n)
        elif g == 3:
            return S_3_n(n)
        else:  # g >= 4:
            return S_g_n(g, n)
    else:  # len(args) > 2:
        raise ValueError('Expected a string or pair of integers.')





''' A module for representing (multi)curves on triangulations. '''

from fractions import Fraction
from collections import Counter, defaultdict
import networkx
import numpy as np

import curver
from curver.kernel.lamination import Lamination, Shortenable  # Special import needed for subclassing.
from curver.kernel.utilities import memoize  # Special import needed for decorating.

class MultiCurve(Lamination):
    ''' A Lamination in which every component is a Curve. '''
    def is_multicurve(self):
        return True
    def is_multiarc(self):
        return False
    def boundary(self):
        return 2*self
    def is_filling(self):
        return False
    
    def encode_twist(self, power=1):
        ''' Return an Encoding of a right Dehn (multi)twist about the components of this multicurve, raised to the given power. '''
        
        h = self.triangulation.id_encoding()
        for curve, multiplicity in self.components().items():
            h = curve.encode_twist(power * multiplicity) * h
        
        return h
    
    def boundary_union(self, other):
        ''' Return \\partial N(self \\cup other). '''
        assert(isinstance(other, Lamination))
        
        crush = self.crush()
        lift = crush.inverse()
        other_prime = crush(other)
        m_prime = other_prime.boundary()
        return lift(m_prime)  # = m.
    
    def fills_with(self, other):
        ''' Return whether self \\cup other fills. '''
        assert(isinstance(other, Lamination))
        
        if any(component.intersection(other) == 0 for component in self.components()):
            return False
        
        crush = self.crush()
        other_prime = crush(other)
        return other_prime.is_filling()
    
    def crush(self):
        ''' Return the crush map associated to this MultiCurve. '''
        
        g = self.triangulation.id_encoding()
        for curve in self.components():
            h = g(curve).crush()  # Map forward under crushes first.
            g = h * g
        
        return g
    
    def is_separating(self):
        ''' Return whether this multicurve separates S.
        
        That is, whether S - self has more components than S. '''
        
        crush = self.crush()
        return len(crush.target_triangulation.components()) > len(self.triangulation.components())
    
    def topological_type(self):
        ''' Return the topological type of this multicurve.
        
        Two multicurves are in the same mapping class group orbit if and only their topological types are equal.
        These are labelled graphs and so equal means 'label isomorphic', so we return a CurvePartitionGraph class that uses networkx.is_isomorphic to determine equality. '''
        
        # We build the CurvePartitionGraph as follows:
        #  1) Crush along all curve components to obtain S'.
        #  2) Create a graph with a vertex for each component of S'.
        #  3) Label each vertex with the topological type of its component.
        #  4) Connect two vertices with an edge for each curve that you crushed along.
        #  5) Label each edge with the multiplicity of the corresponding curve.
        # Then two multicurves are in the same mapping class group orbit iff there is a label-preserving isomorphism between their graphs.
        
        components = self.components()  # The components of this multicurves.
        crush = self.crush()
        lift = crush.inverse()
        triangulation = crush.target_triangulation
        
        graph = networkx.MultiGraph()
        half_edges = defaultdict(list)
        for index, component in enumerate(triangulation.components()):
            vertices = [vertex for vertex in triangulation.vertices if vertex[0] in component]  # The vertices that are in this component.
            V, E = len(vertices), len(component) // 2  # Number of vertices and edges in this component.
            G = (2 - V + E // 3) // 2  # Genus.
            
            graph.add_node(index, genus=G, vertices=V)
            
            for vertex in vertices:
                curve = triangulation.curve_from_cut_sequence(vertex)
                lifted_curve = lift(curve)
                if lifted_curve in components:
                    half_edges[lifted_curve].append(index)
        
        dummy_index = len(graph)
        graph.add_node(dummy_index, genus=0, vertices=0)  # Dummy node for peripheral components.
        
        for curve, nodes in half_edges.items():
            if len(nodes) == 2:
                graph.add_edge(nodes[0], nodes[1], weight=components[curve])
            else:  # len(nodes) == 1:
                graph.add_edge(nodes[0], dummy_index, weight=components[curve])
        
        return curver.kernel.CurvePartitionGraph(self, graph)

class Curve(MultiCurve, Shortenable):
    ''' A MultiCurve with a single component. '''
    @memoize(fast=True)
    def components(self):
        return {self: 1}
    
    def is_short(self):
        if self.is_peripheral(): return True
        
        # Theorem: A non-peripheral curve is short iff either:
        #  - it meets T exactly twice, or
        #  - it meets every edge of T either 0 or 2 times and has one corridor [BellWebb16a].
        num_corridors = len([triangle for triangle in self.triangulation if sum(self(edgy) for edgy in triangle) == 4])
        return self.weight() == 2 or (all(weight in [0, 2] for weight in self) and num_corridors == 1)
    
    def shorten_strategy(self, edge):
        # This relies on the following
        # Lemma: If there are no bipods then self(edge) \in [0,2] for each edge.
        #
        # This follows from the fact that if this is a tripod / monopod in every triangle then
        # connectedness implies that there can't be any part which can't see a vertex.
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        # Low score == bad.
        if not self.triangulation.is_flippable(edge):
            return 0
        
        a, b, c, d, e = self.triangulation.square(edge)
        ai, bi, ci, di, ei = [self(edgy) for edgy in self.triangulation.square(edge)]
        ad, bd, cd, dd, ed = [self.dual_weight(edgy) for edgy in self.triangulation.square(edge)]
        # if ei == 0: return 0
        # return float(2*ei - max(ai + ci, bi + di) + 1 + (0.1 if ad > 0 and bd > 0 else 0)) / self.weight()
        
        if ei == 0:
            return 0
        if max(ai + ci, bi + di) == ei:  # Drops to zero.
            return 1  # Hmmm. We do need this but can we avoid not having 1 as the generic case?
        if ed > 0:
            return 0
        
        if ad > 0 and bd > 0:
            return 0.5
        
        return 0.25
    
    def parallel(self):
        ''' Return an edge that this curve is parallel to.
        
        Note that this is only defined for short, non-peripheral curves. '''
        
        assert(self.is_short() and not self.is_peripheral())
        
        return min([edge for edge in self.triangulation.edges if self(edge) == 0 and self.dual_weight(edge) > 0], key=lambda e: e.label)  # Take the minimum of two.
    
    def is_isolating(self):
        ''' Return if this curve is isolating, that is, if it is non-peripheral and a component of S - self does not contain a puncture.
        
        This curve must be non-peripheral. '''
        
        if self.is_peripheral():
            return False
        
        short, _ = self.shorten()
        return short.weight() > 2
    
    def encode_twist(self, power=1):
        ''' Return an Encoding of a right Dehn twist about this curve, raised to the given power. '''
        
        if self.is_peripheral():  # Boring case.
            return self.triangulation.id_encoding()
        
        short, conjugator = self.shorten()
        
        return conjugator.inverse() * curver.kernel.Twist(short, power).encode() * conjugator
    
    def intersection(self, lamination):
        ''' Return the geometric intersection between self and the given lamination. '''
        
        assert(isinstance(lamination, curver.kernel.Lamination))
        assert(lamination.triangulation == self.triangulation)
        
        if self.is_peripheral():  # Boring case.
            return sum((max(-lamination(edge), 0) + max(-lamination.side_weight(edge), 0)) * self.side_weight(edge) for edge in self.triangulation.edges if self(edge) == 1)
        
        short, conjugator = self.shorten()
        short_lamination = conjugator(lamination)
        
        a = short.parallel()
        v = short.triangulation.vertex_lookup[a.label]  # = self.triangulation.vertex_lookup[~a.label].
        v_edges = curver.kernel.utilities.cyclic_slice(v, a, ~a)  # The set of edges that come out of v from a round to ~a.
        
        around_v = min(max(short_lamination.side_weight(edge), 0) for edge in v_edges)
        out_v = sum(max(-short_lamination.side_weight(edge), 0) for edge in v_edges) + sum(max(-short_lamination(edge), 0) for edge in v_edges[1:])
        # around_v > 0 ==> out_v == 0; out_v > 0 ==> around_v == 0.
        return max(short_lamination(a), 0) - 2 * around_v + out_v
    
    def slope(self, lamination):
        ''' Return the slope of the given lamination about this curve.
        
        This is a Fraction that increases by one each time a right Dehn twist about
        this curve is performed unless -1 <= slope <= 1.
        
        Assumes that this curve and the given lamination intersect.
        This curve must be non-peripheral. '''
        
        assert(not self.is_peripheral())
        
        short, conjugator = self.shorten()
        short_lamination = conjugator(lamination)
        
        denominator = short.intersection(short_lamination)
        if denominator == 0:
            raise curver.AssumptionError('Slope is undefined when curves are disjoint.')
        
        # Get some edges.
        a = short.parallel()
        v = short.triangulation.vertex_lookup[a.label]  # = short.triangulation.vertex_lookup[~a.label].
        _, b, e = short.triangulation.corner_lookup[a.label]
        
        v_edges = curver.kernel.utilities.cyclic_slice(v, a, ~a)  # The set of edges that come out of v from a round to ~a.
        around_v = min(max(short_lamination.side_weight(edge), 0) for edge in v_edges)
        twisting = min(max(short_lamination.side_weight(edge) - around_v, 0) for edge in v_edges[1:-1])
        
        numerator = twisting
        
        sign = -1 if short_lamination.side_weight(a) > around_v or short_lamination.dual_weight(e) < 0 else +1
        
        return Fraction(sign * numerator, denominator) + (1 if sign < 0 and not short.is_isolating() else 0)  # Curver is right biased on non-isolating curves.
    
    def relative_twisting(self, b, c):
        ''' Return the relative twisting number of b about self relative to c.
        
        This is the number of (right) Dehn twists about self that must be applied to b in order to minimise its intersection with c.
        
        Assumes that this curve and the given laminations intersect.
        This curve must be non-peripheral. '''
        
        assert(isinstance(b, curver.kernel.Lamination))
        assert(isinstance(c, curver.kernel.Lamination))
        assert(not self.is_peripheral())
        
        ab = self.intersection(b)
        if ab == 0: raise curver.AssumptionError('Relative slope is undefined when self and b are disjoint.')
        ac = self.intersection(c)  # Faster than c.intersection(a) since we know a is a curve.
        if ac == 0: raise curver.AssumptionError('Relative slope is undefined when self and c are disjoint.')
        bc = b.intersection(c)
        
        f_lower = self.encode_twist(power=-2*bc)(b).intersection(c)  # f(-2*bc).
        f_upper = self.encode_twist(power=2*bc)(b).intersection(c)  # f(2*bc).
        
        return Fraction(f_lower - f_upper, 2*ab*ac)  # No division by zero thanks to our previous checks.
    
    def crush(self):
        ''' Return the crush map associated to this Curve. '''
        
        if self.is_peripheral():  # Boring case.
            return self.triangulation.id_encoding()
        
        short, conjugator = self.shorten()
        
        # Use the following for reference:
        #             #<----------#                #  #-----------#  #
        #            /|     a    ^|               /|  |     a    /  /|
        #           / |         / |              / |  |         /  / |
        #          /  |        /  |             /  |  |        /  /  |
        #         /   |       /   |            /   |  |       /  /   |
        #        /    |b    e/    |   ===>>   /    |  |b   ~b/  /    |
        #       /   ~b|     /~e   |          /    e|  |     /  /~e   |
        #      /      |    /      |         /      |  |    /  /      |
        #     /       |   /       |        /       |  |   /  /       |
        #    /        |  /        |       /        |  |  /  /        |
        #   /         | /         |      /         |  | /  /         |
        #  /          V/          |     /          |  |/  /          |
        # #-----------#-----------#    #-----------#  #  #-----------#
        # Where a is parallel to short.
        
        a = short.parallel()
        a, b, e = short.triangulation.corner_lookup[a.label]
        
        # Build the new triangulation.
        edge_map = dict((edge, curver.kernel.Edge(edge.label)) for edge in short.triangulation.edges)
        # Remap some edges.
        edge_map[e] = curver.kernel.Edge(~b.label)
        edge_map[~b] = curver.kernel.Edge(e.label)
        
        new_triangulation = curver.kernel.Triangulation([curver.kernel.Triangle([edge_map[edgy] for edgy in triangle]) for triangle in short.triangulation])
        
        # Build the lifting matrix back.
        v = short.triangulation.vertex_lookup[a.label]  # = short.triangulation.vertex_lookup[~a.label].
        indices = Counter([edge.index for edge in curver.kernel.utilities.cyclic_slice(v, a, ~a)[1:]])  # The indices that appear walking around v from a to ~a. Note need to exclude the initial a.
        matrix = np.matrix([[indices[j] if i == b.index else 1 if (i == e.index and j == b.index) else 1 if i == j else 0 for i in range(self.zeta)] for j in range(self.zeta)])
        
        crush = curver.kernel.Crush(short.triangulation, new_triangulation, short, matrix).encode()
        return crush * conjugator





import curver

DEFAULT_OBJECT_COLOUR = 'black'
DEFAULT_VERTEX_COLOUR = 'black'
DEFAULT_EDGE_COLOUR = 'black'
DEFAULT_TRIANGLE_COLOUR = 'gray80'
DEFAULT_CURVE_COLOUR = 'grey40'

ARROW_FRAC = 0.55

def dot(a, b):
    return a[0] * b[0] + a[1] * b[1]

def intersection(A, d, B, d2):
    # Find the intersection parameters of A + t d and B + t2 d2
    # Want t & t2 such that:
    #   A.x + t d.x = B.x + t2 d2.x
    #   A.y + t d.y = B.y + t2 d2.y
    # So:
    #   (d.x -d2.x) (t ) = (B.x - A.x)
    #   (d.y -d2.y) (t2)   (B.y - A.y)
    # The inverse of this matrix is:
    #   (-d2.y d2.x)
    #   ( -d.y  d.x) / det
    # where:
    det = d2.cross(d)
    # So:
    t  = ((B.x - A.x) * -d2.y + (B.y - A.y) * d2.x) / det
    t2 = ((B.x - A.x) *  -d.y + (B.y - A.y) *  d.x) / det
    
    return t, t2

def interpolate(A, B, C, r, s):
    # Given points A, B, C and parameters r, s
    # Let X := rB + (1-r)A and
    # Y := sB + (1-s)C
    d = A - B
    d2 = C - B
    
    X = B + r*d
    Y = B + s*d2
    
    centroid = (A + B + C) / 3
    d1a = intersection(X, d.rotate(90), B, centroid - B)[0]
    d1b = intersection(X, d.rotate(90), A, centroid - A)[0]
    t = min([x for x in [d1a, d1b] if x > 0]) / 2
    
    d2a = intersection(Y, d2.rotate(-90), B, centroid - B)[0]
    d2b = intersection(Y, d2.rotate(-90), C, centroid - C)[0]
    t2 = min([x for x in [d2a, d2b] if x > 0]) / 2
    
    P = X + t * d.rotate(90)
    Q = Y + t2 * d2.rotate(-90)
    
    return X, P, Q, Y

class DrawableObject(object):
    def __init__(self, canvas, vertices, options):
        self.options = options
        self.canvas = canvas
        self.vertices = vertices
        self.colour = DEFAULT_OBJECT_COLOUR
        self.drawn = None
    
    def __repr__(self):
        return str(self)
    # Note that this means that CanvasTriangle will NOT have the same convention as AbstractTriangle,
    # there iterating and index accesses return edges.
    def __getitem__(self, index):
        return self.vertices[index % len(self)]
    
    def __iter__(self):
        return iter(self.vertices)
    
    def __len__(self):
        return len(self.vertices)
    
    def set_colour(self, colour=None):
        self.colour = colour
        self.canvas.itemconfig(self.drawn, fill=self.colour)
    
    def centre(self):
        return sum([v.vector for v in self.vertices], curver.application.Vector2(0, 0)) * (1.0 / len(self))
    
    def update(self):
        self.canvas.coords(self.drawn, *[c for v in self for c in v])

class CanvasVertex(DrawableObject):
    def __init__(self, canvas, vector, options):
        super(CanvasVertex, self).__init__(canvas, [self], options)
        self.colour = DEFAULT_VERTEX_COLOUR
        self.vector = vector
        self.drawn = self.canvas.create_oval(
            [p + scale*self.options.dot_size for scale in [-1, 1] for p in self],
            outline=self.colour, fill=self.colour, tag='oval'
            )
    
    def __str__(self):
        return str(self.vector)
    
    def __sub__(self, other):
        return self.vector - other.vector
    
    # We have to redo these manually.
    def __iter__(self):
        return iter(self.vector)
    
    def __contains__(self, point):
        return all(abs(c - v) < self.options.epsilon for c, v in zip(point, self))
    
    def update(self):
        self.canvas.coords(self.drawn, *[p + scale*self.options.dot_size for scale in [-1, 1] for p in self])

class CanvasEdge(DrawableObject):
    def __init__(self, canvas, vertices, label, colour, options):
        super(CanvasEdge, self).__init__(canvas, vertices, options)
        self.label = label
        self.colour = DEFAULT_EDGE_COLOUR if colour is None else colour
        m = (1-ARROW_FRAC)*self.vertices[0].vector + ARROW_FRAC*self.vertices[1].vector
        self.drawn = [  # Need two lines really so arrows work correctly.
            self.canvas.create_line(
                [c for v in [self.vertices[0], m] for c in v],
                width=self.options.line_size,
                fill=self.colour,
                tags=['line', 'line_start'],
                arrowshape=self.options.arrow_shape
            ),
            self.canvas.create_line(
                [c for v in self.vertices for c in v],
                width=self.options.line_size,
                fill=self.colour,
                tags=['line', 'line_end'],
                arrowshape=self.options.arrow_shape
            )
            ]
        
        self.in_triangles = []
    
    def __str__(self):
        return str(self.vertices)
    
    def hide(self, hide=False):
        for drawn in self.drawn:
            self.canvas.itemconfig(drawn, state='hidden' if hide else 'normal')
    
    def update(self):
        m = (1-ARROW_FRAC)*self.vertices[0].vector + ARROW_FRAC*self.vertices[1].vector
        self.canvas.coords(self.drawn[0], *[c for v in [self.vertices[0], m] for c in v])
        self.canvas.coords(self.drawn[1], *[c for v in self.vertices for c in v])

class CanvasTriangle(DrawableObject):
    def __init__(self, canvas, edges, options):
        super(CanvasTriangle, self).__init__(canvas, list(set(v for e in edges for v in e)), options)
        self.colour = DEFAULT_TRIANGLE_COLOUR
        self.edges = edges
        
        # We reorder the vertices to guarantee that the vertices are cyclically ordered anticlockwise in the plane.
        if (self[1] - self[0]).cross(self[2] - self[0]) > 0: self.vertices = [self[0], self[2], self[1]]
        # Now we reorder the edges such that edges[i] does not meet vertices[i].
        self.edges = [edge for vertex in self for edge in self.edges if vertex not in edge.vertices]
        
        # And check to make sure everyone made it through alive.
        assert(len(self.edges) == 3)
        assert(self[0] != self[1] and self[1] != self[2] and self[2] != self[0])
        assert(self.edges[0] != self.edges[1] and self.edges[1] != self.edges[2] and self.edges[2] != self.edges[0])
        
        self.drawn = self.canvas.create_polygon([c for v in self for c in v], fill=self.colour, tag='polygon')
        # Add this triangle to each edge involved.
        for edge in self.edges:
            edge.in_triangles.append(self)
    
    def __str__(self):
        return str(self.vertices)

class CurveComponent(DrawableObject):
    def __init__(self, canvas, vertices, options, thin=True, smooth=False):
        super(CurveComponent, self).__init__(canvas, vertices, options)
        self.colour = DEFAULT_CURVE_COLOUR
        if thin:
            self.drawn = self.canvas.create_line(
                [c for v in self.vertices for c in v],
                width=self.options.line_size,
                fill=self.colour,
                tag='curve',
                smooth=smooth,
                splinesteps=50
                )
        else:
            self.drawn = self.canvas.create_polygon(
                [c for v in self.vertices for c in v],
                fill=self.colour,
                tag='curve',
                outline=self.colour,
                smooth=smooth,
                splinesteps=50
                )





''' Curver is a program for computations in the curve complex. '''

import curver.kernel
from curver.load import load
from numbers import Integral as IntegerType

import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import pkg_resources  # Suppress 'UserWarning: Module curver was already imported from ...'
    __version__ = pkg_resources.require('curver')[0].version

# Set up really short names for the most commonly used classes and functions by users.
create_triangulation = curver.kernel.create_triangulation
triangulation_from_sig = curver.kernel.triangulation_from_sig

AbortError = curver.kernel.AbortError
AssumptionError = curver.kernel.AssumptionError

def show(*items):
    import curver.application
    curver.application.start(*items)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestMCG(unittest.TestCase):
    @given(strategies.mcgs())
    @settings(max_examples=1)
    def test_pickle(self, mcg):
        self.assertEqual(mcg, pickle.loads(pickle.dumps(mcg)))
    
    @given(st.data())
    def test_curve_intersection(self, data):
        mcg = data.draw(strategies.mcgs())
        name1 = data.draw(st.sampled_from(sorted(mcg.curves)))
        name2 = data.draw(st.sampled_from(sorted(mcg.curves)))
        curve1 = mcg.curves[name1]
        curve2 = mcg.curves[name2]
        intersection = curve1.intersection(curve2)
        self.assertGreaterEqual(intersection, 0)
        self.assertLessEqual(intersection, 1)
    
    @given(st.data())
    def test_arc_intersection(self, data):
        mcg = data.draw(strategies.mcgs())
        name1 = data.draw(st.sampled_from(sorted(mcg.arcs)))
        name2 = data.draw(st.sampled_from(sorted(mcg.arcs)))
        arc1 = mcg.arcs[name1]
        arc2 = mcg.arcs[name2]
        self.assertEqual(arc1.intersection(arc2), 0)
    
    @given(st.data())
    @settings(max_examples=50)
    def test_curve_relation(self, data):
        mcg = data.draw(strategies.mcgs())
        name1 = data.draw(st.sampled_from(sorted(mcg.curves)))
        name2 = data.draw(st.sampled_from(sorted(mcg.curves)))
        curve1 = mcg.curves[name1]
        curve2 = mcg.curves[name2]
        intersection = curve1.intersection(curve2)
        self.assertTrue(
            (intersection == 0 and mcg(name1 + name2) == mcg(name2 + name1)) or \
            (intersection == 1 and mcg(name1 + name2 + name1) == mcg(name2 + name1 + name2)) or \
            intersection >= 2
            )

    @given(st.data())
    @settings(max_examples=50)
    def test_arc_relation(self, data):
        mcg = data.draw(strategies.mcgs())
        distinct_end_arcs = sorted(name for name, arc in mcg.arcs.items() if arc.connects_distinct_vertices)
        name1 = data.draw(st.sampled_from(distinct_end_arcs))
        name2 = data.draw(st.sampled_from(distinct_end_arcs))  # Hmm, should we check arc1.intersection(arc2) == 0?
        arc1 = mcg.arcs[name1]
        arc2 = mcg.arcs[name2]
        num_distinct_vertices = len(set(arc1.vertices() + arc2.vertices()))
        
        self.assertTrue(
            (num_distinct_vertices == 4 and mcg(name1 + name2) == mcg(name2 + name1)) or \
            (num_distinct_vertices == 3 and mcg(name1 + name2 + name1) == mcg(name2 + name1 + name2)) or \
            (num_distinct_vertices == 2)
            )
    
    @given(st.data())
    @settings(max_examples=2)
    def test_expand_word(self, data):
        mcg = data.draw(strategies.mcgs())
        word1 = mcg.random_word(data.draw(st.integers(min_value=0, max_value=10)))
        word2 = mcg.random_word(data.draw(st.integers(min_value=0, max_value=10)))
        power = data.draw(st.integers(min_value=-10, max_value=10))
        
        self.assertEqual(mcg(word1 + word2), mcg(word1) * mcg(word2))
        self.assertEqual(mcg('(%s)^%d' % (word1, power)), mcg(word1)**power)





''' A module for representing partition graphs. '''

from itertools import permutations, combinations_with_replacement
import networkx
import curver

class CurvePartitionGraph(object):
    ''' This represents a partition graph of a MultiCurve. '''
    def __init__(self, multicurve, graph):
        self.multicurve = multicurve
        self.graph = graph
    
    def __str__(self):
        G = self.graph
        N = G.nodes(data=True)
        
        return min(','.join('%d:%d' % (N[p]['genus'], N[p]['vertices']) for p in perm) + '-' + ','.join([':'.join(str(x) for x in sorted(d['weight'] for d in G.get_edge_data(i, j, default={}).values())) for i, j in combinations_with_replacement(perm, r=2)]) for perm in permutations(range(len(G))))
    
    def __repr__(self):
        return str(self)
    
    def __eq__(self, other):
        node_match = lambda v1, v2: v1['genus'] == v2['genus'] and v1['vertices'] == v2['vertices']
        edge_match = lambda e1, e2: sorted(e['weight'] for e in e1.values()) == sorted(e['weight'] for e in e2.values())
        return networkx.is_isomorphic(self.graph, other.graph, node_match=node_match, edge_match=edge_match)
    def __ne__(self, other):
        return not (self == other)
    
    def __hash__(self):
        return hash(tuple(sorted((d['genus'], d['vertices']) for _, d in self.graph.nodes(data=True))) + tuple(sorted([e['weight'] for _, _, e in self.graph.edges(data=True)])))





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestLamination(unittest.TestCase):
    # @pytest.mark.slow
    @given(strategies.laminations())
    def test_pickle(self, lamination):
        self.assertEqual(lamination, pickle.loads(pickle.dumps(lamination)))
    
    # @pytest.mark.slow
    @given(st.data())
    def test_hash(self, data):
        lamination1 = data.draw(strategies.laminations())
        lamination2 = data.draw(strategies.laminations(lamination1.triangulation))
        self.assertTrue(hash(lamination1) != hash(lamination2) or lamination1 == lamination2)
    
    # @pytest.mark.slow
    @given(st.data())
    def test_orientation(self, data):
        lamination = data.draw(strategies.laminations())
        edge = data.draw(st.sampled_from(lamination.triangulation.edges))
        self.assertEqual(lamination(edge), lamination(~edge))

    # @pytest.mark.slow
    @given(strategies.laminations())
    @settings(max_examples=20)
    def test_components(self, lamination):
        self.assertEqual(lamination.triangulation.sum([multiplicity * component for component, multiplicity in lamination.components().items()]), lamination)
        self.assertEqual(lamination.triangulation.disjoint_sum([multiplicity * component for component, multiplicity in lamination.components().items()]), lamination)
        for component in lamination.components():
            self.assertEqual(component.intersection(component), 0)
    
    # @pytest.mark.slow
    @given(st.data())
    @settings(max_examples=20)
    def test_components_image(self, data):
        lamination = data.draw(strategies.laminations())
        encoding = data.draw(strategies.encodings(lamination.triangulation))
        self.assertEqual(set(encoding(lamination).components()), {encoding(component) for component in lamination.components()})
    
    # @pytest.mark.slow
    @given(st.data())
    @settings(max_examples=10)
    def test_intersection(self, data):
        lamination1 = data.draw(strategies.laminations())
        lamination2 = data.draw(strategies.laminations(lamination1.triangulation))
        encoding = data.draw(strategies.encodings(lamination1.triangulation))
        self.assertGreaterEqual(lamination1.intersection(lamination2), 0)
        self.assertEqual(lamination1.intersection(lamination2), lamination2.intersection(lamination1))
        self.assertEqual(lamination1.intersection(lamination2), encoding(lamination1).intersection(encoding(lamination2)))





from hypothesis import given
import hypothesis.strategies as st
from hypothesis.stateful import RuleBasedStateMachine, rule
import pytest
from string import ascii_lowercase
import unittest

import curver
import strategies

class UnionFindRules(RuleBasedStateMachine):
    def __init__(self):
        super(UnionFindRules, self).__init__()
        self.initialize([])
    
    @rule(items=st.sets(elements=st.integers()))
    def initialize(self, items):
        self.__items = list(items)
        self.__union_find = curver.kernel.utilities.UnionFind(items)
    
    @rule(data=st.data())
    def find(self, data):
        a = data.draw(st.sampled_from(self.__items))
        self.__union_find(a)
    
    @rule(data=st.data())
    def iterate(self, data):
        assert set(self.__items) == set(sum(self.__union_find, []))
    
    @rule(data=st.data())
    def union2(self, data):
        a = data.draw(st.sampled_from(self.__items))
        b = data.draw(st.sampled_from(self.__items))
        self.__union_find.union2(a, b)
        assert self.__union_find(a) == self.__union_find(b)
    
    @rule(data=st.data())
    def union(self, data):
        items = data.draw(st.lists(elements=st.sampled_from(self.__items)))
        self.__union_find.union(items)
        for item in items:
            assert self.__union_find(item) == self.__union_find(items[0])

TestUnionFind = UnionFindRules.TestCase

class TestMisc(unittest.TestCase):
    @given(st.integers(min_value=0))
    def test_b64(self, integer):
        strn = curver.kernel.utilities.b64encode(integer)
        self.assertEqual(curver.kernel.utilities.b64decode(strn), integer)

    @given(st.integers(min_value=0, max_value=1000), st.sets(elements=st.text(ascii_lowercase)))
    def test_string_generator(self, n, skip):
        strns = curver.kernel.utilities.string_generator(n, skip)
        self.assertEqual(len(strns), n)
        self.assertFalse(skip.intersection(strns))
    
    @given(st.data())
    def test_maximum(self, data):
        bound = data.draw(st.integers())
        integers = data.draw(st.lists(elements=st.integers(max_value=bound), min_size=1))
        value = curver.kernel.utilities.maximum(integers, upper_bound=bound)
        self.assertEqual(value, min(max(integers), bound))

